from flask import make_response

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import functools
import hashlib
import hmac
import logging
import os
import threading
import time
from datetime import datetime, timedelta, timezone
from typing import Any, Union

import flask
import flask.json
import flask.json.provider
import flask.sessions
import flask.templating
import flask_assets
import flask_login
import netaddr
import tornado.web
import webassets.updater
import webassets.utils
from cachelib import BaseCache
from flask import current_app
from flask_login import COOKIE_NAME as REMEMBER_COOKIE_NAME
from flask_login.utils import decode_cookie, encode_cookie
from werkzeug.local import LocalProxy
from werkzeug.utils import cached_property

import octoprint.access.users
import octoprint.plugin
import octoprint.server
import octoprint.vendor.flask_principal as flask_principal
from octoprint.access import auth_log
from octoprint.events import Events, eventManager
from octoprint.settings import settings
from octoprint.util import DefaultOrderedDict, deprecated, yaml
from octoprint.util.json import JsonEncoding
from octoprint.util.net import is_lan_address
from octoprint.util.tz import UTC_TZ, is_timezone_aware

# ~~ monkey patching


def enable_additional_translations(default_locale="en", additional_folders=None):
    import os

    import flask_babel
    from babel import Locale, support

    if additional_folders is None:
        additional_folders = []

    logger = logging.getLogger(__name__)

    def fixed_list_translations(self):
        """Returns a list of all the locales translations exist for.  The
        list returned will be filled with actual locale objects and not just
        strings.
        """

        def list_translations(dirname):
            if not os.path.isdir(dirname):
                return []
            result = []
            for entry in os.scandir(dirname):
                locale_dir = os.path.join(entry.path, "LC_MESSAGES")
                if not os.path.isdir(locale_dir):
                    continue
                if any(filter(lambda x: x.name.endswith(".mo"), os.scandir(locale_dir))):
                    result.append(Locale.parse(entry.name))
            return result

        dirs = additional_folders + [os.path.join(self.app.root_path, "translations")]

        # translations from plugins
        plugins = octoprint.plugin.plugin_manager().enabled_plugins
        for plugin in plugins.values():
            plugin_translation_dir = os.path.join(plugin.location, "translations")
            if not os.path.isdir(plugin_translation_dir):
                continue
            dirs.append(plugin_translation_dir)

        result = {Locale.parse(default_locale)}

        for dir in dirs:
            result.update(list_translations(dir))
        return list(result)

    def fixed_get_translations():
        """Returns the correct gettext translations that should be used for
        this request.  This will never fail and return a dummy translation
        object if used outside of the request or if a translation cannot be
        found.
        """
        if flask.g is None:
            return None
        translations = getattr(flask.g, "babel_translations", None)
        if translations is None:
            locale = flask_babel.get_locale()
            translations = support.Translations()

            if str(locale) != default_locale:
                # plugin translations
                plugins = octoprint.plugin.plugin_manager().enabled_plugins
                for name, plugin in plugins.items():
                    dirs = list(
                        map(
                            lambda x: os.path.join(x, "_plugins", name),
                            additional_folders,
                        )
                    ) + [os.path.join(plugin.location, "translations")]
                    for dirname in dirs:
                        if not os.path.isdir(dirname):
                            continue

                        try:
                            plugin_translations = support.Translations.load(
                                dirname, [locale]
                            )
                        except Exception:
                            logger.exception(
                                f"Error while trying to load translations "
                                f"for plugin {name}"
                            )
                        else:
                            if isinstance(plugin_translations, support.Translations):
                                translations = translations.merge(plugin_translations)
                                logger.debug(
                                    f"Using translation plugin folder {dirname} from "
                                    f"plugin {name} for locale {locale}"
                                )
                                break
                    else:
                        logger.debug(
                            f"No translations for locale {locale} " f"from plugin {name}"
                        )

                # core translations
                dirs = additional_folders + [
                    os.path.join(flask.current_app.root_path, "translations")
                ]
                for dirname in dirs:
                    core_translations = support.Translations.load(dirname, [locale])
                    if isinstance(core_translations, support.Translations):
                        logger.debug(
                            f"Using translation core folder {dirname} "
                            f"for locale {locale}"
                        )
                        break
                else:
                    logger.debug(f"No translations for locale {locale} in core folders")
                translations = translations.merge(core_translations)

            flask.g.babel_translations = translations
        return translations

    flask_babel.Babel.list_translations = fixed_list_translations
    flask_babel.get_translations = fixed_get_translations


def fix_webassets_filtertool():
    from webassets.merge import FilterTool, MemoryHunk, log

    error_logger = logging.getLogger(__name__ + ".fix_webassets_filtertool")

    def fixed_wrap_cache(self, key, func):
        """Return cache value ``key``, or run ``func``."""
        if self.cache:
            if not self.no_cache_read:
                log.debug("Checking cache for key %s", key)
                content = self.cache.get(key)
                if content not in (False, None):
                    log.debug("Using cached result for %s", key)
                    return MemoryHunk(content)

        try:
            content = func().getvalue()
            if self.cache:
                try:
                    log.debug(
                        "Storing result in cache with key %s",
                        key,
                    )
                    self.cache.set(key, content)
                except Exception:
                    error_logger.exception(
                        "Got an exception while trying to save file to cache, not caching"
                    )
            return MemoryHunk(content)
        except Exception:
            error_logger.exception(
                "Got an exception while trying to apply filter, ignoring file"
            )
            return MemoryHunk("")

    FilterTool._wrap_cache = fixed_wrap_cache


def fix_webassets_convert_item_to_flask_url():
    import flask_assets

    def fixed_convert_item_to_flask_url(self, ctx, item, filepath=None):
        from flask import url_for

        directory, rel_path, endpoint = self.split_prefix(ctx, item)

        if filepath is not None:
            filename = filepath[len(directory) + 1 :]
        else:
            filename = rel_path

        flask_ctx = None
        if not flask.has_request_context():  # fixed, was _request_ctx.top
            flask_ctx = ctx.environment._app.test_request_context()
            flask_ctx.push()
        try:
            url = url_for(endpoint, filename=filename)
            # In some cases, url will be an absolute url with a scheme and hostname.
            # (for example, when using werkzeug's host matching).
            # In general, url_for() will return a http url. During assets build, we
            # we don't know yet if the assets will be served over http, https or both.
            # Let's use // instead. url_for takes a _scheme argument, but only together
            # with external=True, which we do not want to force every time. Further,
            # this _scheme argument is not able to render // - it always forces a colon.
            if url and url.startswith("http:"):
                url = url[5:]
            return url
        finally:
            if flask_ctx:
                flask_ctx.pop()

    flask_assets.FlaskResolver.convert_item_to_flask_url = fixed_convert_item_to_flask_url


# ~~ WSGI environment wrapper for reverse proxying


class ReverseProxiedEnvironment:
    @staticmethod
    def to_header_candidates(values):
        if values is None:
            return []
        if not isinstance(values, (list, tuple)):
            values = [values]
        to_wsgi_format = lambda header: "HTTP_" + header.upper().replace("-", "_")
        return list(map(to_wsgi_format, values))

    @staticmethod
    def valid_ip(address):
        import netaddr

        try:
            netaddr.IPAddress(address)
            return True
        except Exception:
            return False

    def __init__(
        self,
        header_prefix=None,
        header_scheme=None,
        header_host=None,
        header_server=None,
        header_port=None,
        prefix=None,
        scheme=None,
        host=None,
        server=None,
        port=None,
    ):
        # sensible defaults
        if header_prefix is None:
            header_prefix = ["x-script-name", "x-forwarded-prefix"]
        if header_scheme is None:
            header_scheme = ["x-forwarded-proto", "x-scheme"]
        if header_host is None:
            header_host = ["x-forwarded-host"]
        if header_server is None:
            header_server = ["x-forwarded-server"]
        if header_port is None:
            header_port = ["x-forwarded-port"]

        # header candidates
        self._headers_prefix = self.to_header_candidates(header_prefix)
        self._headers_scheme = self.to_header_candidates(header_scheme)
        self._headers_host = self.to_header_candidates(header_host)
        self._headers_server = self.to_header_candidates(header_server)
        self._headers_port = self.to_header_candidates(header_port)

        # fallback prefix & scheme & host from config
        self._fallback_prefix = prefix
        self._fallback_scheme = scheme
        self._fallback_host = host
        self._fallback_server = server
        self._fallback_port = port

    def __call__(self, environ):
        def retrieve_header(header_type):
            candidates = getattr(self, "_headers_" + header_type, [])
            fallback = getattr(self, "_fallback_" + header_type, None)

            for candidate in candidates:
                value = environ.get(candidate, None)
                if value is not None:
                    return value
            else:
                return fallback

        def host_to_server_and_port(host, scheme):
            if host is None:
                return None, None

            default_port = "443" if scheme == "https" else "80"
            host = host.strip()

            if ":" in host:
                # we might have an ipv6 address here, or a port, or both

                if host[0] == "[":
                    # that looks like an ipv6 address with port, e.g. [fec1::1]:80
                    address_end = host.find("]")
                    if address_end == -1:
                        # no ], that looks like a seriously broken address
                        return None, None

                    # extract server ip, skip enclosing [ and ]
                    server = host[1:address_end]
                    tail = host[address_end + 1 :]

                    # now check if there's also a port
                    if len(tail) and tail[0] == ":":
                        # port included as well
                        port = tail[1:]
                    else:
                        # no port, use default one
                        port = default_port

                elif self.__class__.valid_ip(host):
                    # ipv6 address without port
                    server = host
                    port = default_port

                else:
                    # ipv4 address with port
                    server, port = host.rsplit(":", 1)

            else:
                server = host
                port = default_port

            return server, port

        # determine prefix
        prefix = retrieve_header("prefix")
        if prefix is not None:
            environ["SCRIPT_NAME"] = prefix
            path_info = environ["PATH_INFO"]
            if path_info.startswith(prefix):
                environ["PATH_INFO"] = path_info[len(prefix) :]

        # determine scheme
        scheme = retrieve_header("scheme")
        if scheme is not None and "," in scheme:
            # Scheme might be something like "https,https" if doubly-reverse-proxied
            # without stripping original scheme header first, make sure to only use
            # the first entry in such a case. See #1391.
            scheme, _ = map(lambda x: x.strip(), scheme.split(",", 1))
        if scheme is not None:
            environ["wsgi.url_scheme"] = scheme

        # determine host
        url_scheme = environ["wsgi.url_scheme"]
        host = retrieve_header("host")
        if host is not None:
            # if we have a host, we take server_name and server_port from it
            server, port = host_to_server_and_port(host, url_scheme)
            environ["HTTP_HOST"] = host
            environ["SERVER_NAME"] = server
            environ["SERVER_PORT"] = port

        elif environ.get("HTTP_HOST", None) is not None:
            # if we have a Host header, we use that and make sure our server name and port properties match it
            host = environ["HTTP_HOST"]
            server, port = host_to_server_and_port(host, url_scheme)
            environ["SERVER_NAME"] = server
            environ["SERVER_PORT"] = port

        else:
            # else we take a look at the server and port headers and if we have
            # something there we derive the host from it

            # determine server - should usually not be used
            server = retrieve_header("server")
            if server is not None:
                environ["SERVER_NAME"] = server

            # determine port - should usually not be used
            port = retrieve_header("port")
            if port is not None:
                environ["SERVER_PORT"] = port

            # reconstruct host header
            if (
                url_scheme == "http"
                and environ["SERVER_PORT"] == "80"
                or url_scheme == "https"
                and environ["SERVER_PORT"] == "443"
            ):
                # default port for scheme, can be skipped
                environ["HTTP_HOST"] = environ["SERVER_NAME"]
            else:
                server_name_component = environ["SERVER_NAME"]
                if ":" in server_name_component and self.__class__.valid_ip(
                    server_name_component
                ):
                    # this is an ipv6 address, we need to wrap that in [ and ] before appending the port
                    server_name_component = "[" + server_name_component + "]"

                environ["HTTP_HOST"] = (
                    server_name_component + ":" + environ["SERVER_PORT"]
                )

        # call wrapped app with rewritten environment
        return environ


# ~~ request and response versions


def encode_remember_me_cookie(value):
    from octoprint.server import userManager

    name = value.split("|")[0]
    try:
        remember_key = userManager.signature_key_for_user(
            name, current_app.config["SECRET_KEY"]
        )
        timestamp = datetime.utcnow().timestamp()
        return encode_cookie(f"{name}|{timestamp}", key=remember_key)
    except Exception:
        pass

    return ""


def decode_remember_me_cookie(value):
    from octoprint.server import userManager

    parts = value.split("|")
    if len(parts) == 3:
        name, created, _ = parts

        try:
            # valid signature?
            signature_key = userManager.signature_key_for_user(
                name, current_app.config["SECRET_KEY"]
            )
            cookie = decode_cookie(value, key=signature_key)
            if cookie:
                # still valid?
                if (
                    datetime.fromtimestamp(float(created))
                    + timedelta(seconds=current_app.config["REMEMBER_COOKIE_DURATION"])
                    > datetime.utcnow()
                ):
                    return encode_cookie(name)
        except Exception:
            pass

    raise ValueError("Invalid remember me cookie")


def get_cookie_suffix(request):
    """
    Request specific suffix for set and read cookies

    We need this because cookies are not port-specific and we don't want to overwrite our
    session and other cookies from one OctoPrint instance on our machine with those of another
    one who happens to listen on the same address albeit a different port or script root.
    """
    result = "_P" + request.server_port
    if request.script_root:
        return result + "_R" + request.script_root.replace("/", "|")
    return result


class OctoPrintFlaskRequest(flask.Request):
    environment_wrapper = staticmethod(lambda x: x)

    def __init__(self, environ, *args, **kwargs):
        # apply environment wrapper to provided WSGI environment
        flask.Request.__init__(self, self.environment_wrapper(environ), *args, **kwargs)

    @cached_property
    def cookies(self):
        # strip cookie_suffix from all cookies in the request, return result
        cookies = flask.Request.cookies.__get__(self)

        result = {}
        desuffixed = {}
        for key, value in cookies.items():

            def process_value(k, v):
                if k == current_app.config.get(
                    "REMEMBER_COOKIE_NAME", REMEMBER_COOKIE_NAME
                ):
                    return decode_remember_me_cookie(v)
                return v

            try:
                if key.endswith(self.cookie_suffix):
                    key = key[: -len(self.cookie_suffix)]
                    desuffixed[key] = process_value(key, value)
                else:
                    result[key] = process_value(key, value)
            except ValueError:
                # ignore broken cookies
                pass

        result.update(desuffixed)
        return result

    @cached_property
    def server_name(self):
        """Short cut to the request's server name header"""
        return self.environ.get("SERVER_NAME")

    @cached_property
    def server_port(self):
        """Short cut to the request's server port header"""
        return self.environ.get("SERVER_PORT")

    @cached_property
    def cookie_suffix(self):
        return get_cookie_suffix(self)


class OctoPrintFlaskResponse(flask.Response):
    def set_cookie(self, key, value="", *args, **kwargs):
        # restrict cookie path to script root
        kwargs["path"] = flask.request.script_root + kwargs.get("path", "/")

        # set same-site header
        samesite = settings().get(["server", "cookies", "samesite"])
        if samesite is not None:
            samesite = samesite.lower()
        if samesite == "none":
            # Must be string "None"
            samesite = "None"
        if samesite not in ("None", "strict", "lax"):
            # If NoneType, the cookie is not set
            samesite = None
        kwargs["samesite"] = samesite

        # set secure if necessary
        kwargs["secure"] = flask.request.environ.get(
            "wsgi.url_scheme"
        ) == "https" or settings().getBoolean(["server", "cookies", "secure"])

        # tie account properties to remember me cookie (e.g. current password hash)
        if key == current_app.config.get("REMEMBER_COOKIE_NAME", REMEMBER_COOKIE_NAME):
            value = encode_remember_me_cookie(value)

        # add request specific cookie suffix to name
        flask.Response.set_cookie(
            self, key + flask.request.cookie_suffix, value=value, *args, **kwargs
        )

    def delete_cookie(self, key, path="/", domain=None):
        flask.Response.delete_cookie(self, key, path=path, domain=domain)

        # we also still might have a cookie left over from before we started prefixing, delete that manually
        # without any pre processing (no path prefix, no key suffix)
        flask.Response.set_cookie(
            self, key, expires=0, max_age=0, path=path, domain=domain
        )


class OctoPrintSessionInterface(flask.sessions.SecureCookieSessionInterface):
    def should_set_cookie(self, app, session):
        return flask.request.endpoint != "static"

    def save_session(self, app, session, response):
        if flask.g.get("login_via_apikey", False):
            return
        return super().save_session(app, session, response)


# ~~ jinja environment


class PrefixAwareJinjaEnvironment(flask.templating.Environment):
    def __init__(self, *args, **kwargs):
        flask.templating.Environment.__init__(self, *args, **kwargs)
        self.prefix_loader = None
        self._cached_templates = {}

    def join_path(self, template, parent):
        if parent and "/" in parent:
            prefix, _ = parent.split("/", 1)
            if template in self._templates_for_prefix(prefix) and not template.startswith(
                prefix + "/"
            ):
                return prefix + "/" + template

        return template

    def _templates_for_prefix(self, prefix):
        if prefix in self._cached_templates:
            return self._cached_templates[prefix]

        templates = []
        if prefix in self.prefix_loader.mapping:
            templates = self.prefix_loader.mapping[prefix].list_templates()
        self._cached_templates[prefix] = templates
        return templates


# ~~ passive login helper

_cached_local_networks = None


def _local_networks():
    global _cached_local_networks

    if _cached_local_networks is None:
        logger = logging.getLogger(__name__)
        local_networks = netaddr.IPSet([])
        for entry in settings().get(["accessControl", "localNetworks"]):
            try:
                network = netaddr.IPNetwork(entry)
            except Exception:
                logger.warning(
                    "Invalid network definition configured in localNetworks: {}".format(
                        entry
                    )
                )
                continue

            local_networks.add(network)
            logger.debug(f"Added network {network} to localNetworks")

            if network.version == 4:
                network_v6 = network.ipv6()
                local_networks.add(network_v6)
                logger.debug(
                    "Also added v6 representation of v4 network {} = {} to localNetworks".format(
                        network, network_v6
                    )
                )

        _cached_local_networks = local_networks

    return _cached_local_networks


def passive_login():
    logger = logging.getLogger(__name__)

    user = flask_login.current_user

    remote_address = get_remote_address(flask.request)
    ip_check_enabled = settings().getBoolean(["server", "ipCheck", "enabled"])
    ip_check_trusted = settings().get(["server", "ipCheck", "trustedSubnets"])

    if isinstance(user, LocalProxy):
        # noinspection PyProtectedMember
        user = user._get_current_object()

    def login(u):
        # login known user
        if not u.is_anonymous:
            if not flask.g.identity or not flask.g.identity.id:
                # the user was just now found
                login_mechanism = octoprint.server.util.LoginMechanism.to_log(
                    flask.session.get("login_mechanism", "unknown")
                )
                auth_log(
                    f"Logging in user {u.get_id()} from {remote_address} via {login_mechanism}"
                )
            u = octoprint.server.userManager.login_user(u)

        flask_login.login_user(u)
        flask_principal.identity_changed.send(
            flask.current_app._get_current_object(),
            identity=flask_principal.Identity(u.get_id()),
        )
        if hasattr(u, "session"):
            flask.session["usersession.id"] = u.session
            flask.session["usersession.signature"] = session_signature(
                u.get_id(), u.session
            )
        flask.g.user = u

        eventManager().fire(Events.USER_LOGGED_IN, payload={"username": u.get_id()})

        return u

    def determine_user(u):
        if not u.is_anonymous and u.is_active:
            # known active user
            logger.info(f"Passively logging in user {u.get_id()} from {remote_address}")

        elif (
            settings().getBoolean(["accessControl", "autologinLocal"])
            and settings().get(["accessControl", "autologinAs"]) is not None
            and settings().get(["accessControl", "localNetworks"]) is not None
            and "active_logout" not in flask.request.cookies
            and remote_address
        ):
            # attempt local autologin
            autologin_as = settings().get(["accessControl", "autologinAs"])
            local_networks = _local_networks()
            logger.debug(
                "Checking if remote address {} is in localNetworks ({!r})".format(
                    remote_address, local_networks
                )
            )

            try:
                if netaddr.IPAddress(remote_address) in local_networks:
                    autologin_user = octoprint.server.userManager.find_user(autologin_as)
                    if autologin_user is not None and autologin_user.is_active:
                        logger.info(
                            f"Logging in user {autologin_as} from {remote_address} via autologin"
                        )
                        flask.session[
                            "login_mechanism"
                        ] = octoprint.server.util.LoginMechanism.AUTOLOGIN
                        flask.session["credentials_seen"] = False
                        return autologin_user
            except Exception:
                logger.exception(
                    "Could not autologin user {} from {} for networks {}".format(
                        autologin_as, remote_address, local_networks
                    )
                )

        if not u.is_active:
            # inactive user, switch to anonymous
            u = octoprint.server.userManager.anonymous_user_factory()

        return u

    user = login(determine_user(user))
    response = user.as_dict()
    response["_is_external_client"] = ip_check_enabled and not is_lan_address(
        remote_address, additional_private=ip_check_trusted
    )
    if flask.session.get("login_mechanism") is not None:
        response["_login_mechanism"] = flask.session.get("login_mechanism")
    response["_credentials_seen"] = to_api_credentials_seen(
        flask.session.get("credentials_seen", False)
    )
    return flask.jsonify(response)


def to_api_credentials_seen(credentials_seen):
    if not credentials_seen:
        return False

    return (
        datetime.fromtimestamp(credentials_seen, tz=timezone.utc)
        .replace(microsecond=0)
        .isoformat()
    )


# ~~ rate limiting helper


def limit(*args, **kwargs):
    if octoprint.server.limiter:
        return octoprint.server.limiter.limit(*args, **kwargs)
    else:

        def decorator(f):
            @functools.wraps(f)
            def decorated_function(*args, **kwargs):
                return f(*args, **kwargs)

            return decorated_function

        return decorator


# ~~ cache decorator for cacheable views


class LessSimpleCache(BaseCache):
    """
    Slightly improved version of :class:`SimpleCache`.

    Setting ``default_timeout`` or ``timeout`` to ``-1`` will have no timeout be applied at all.
    """

    def __init__(self, threshold=500, default_timeout=300):
        BaseCache.__init__(self, default_timeout=default_timeout)
        self._mutex = threading.RLock()
        self._cache = {}
        self._bypassed = set()
        self.clear = self._cache.clear
        self._threshold = threshold

    def _prune(self):
        if self.over_threshold():
            now = time.time()
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if expires is not None and expires <= now or idx % 3 == 0:
                    with self._mutex:
                        self._cache.pop(key, None)

    def get(self, key):
        import pickle

        now = time.time()
        with self._mutex:
            expires, value = self._cache.get(key, (0, None))
        if expires is None or expires > now:
            return pickle.loads(value)

    def set(self, key, value, timeout=None):
        import pickle

        with self._mutex:
            self._prune()
            self._cache[key] = (
                self.calculate_timeout(timeout=timeout),
                pickle.dumps(value, pickle.HIGHEST_PROTOCOL),
            )
            if key in self._bypassed:
                self._bypassed.remove(key)

    def add(self, key, value, timeout=None):
        with self._mutex:
            self.set(key, value, timeout=None)
            self._cache.setdefault(key, self._cache[key])

    def delete(self, key):
        with self._mutex:
            self._cache.pop(key, None)

    def calculate_timeout(self, timeout=None):
        if timeout is None:
            timeout = self.default_timeout
        if timeout == -1:
            return None
        return time.time() + timeout

    def over_threshold(self):
        if self._threshold is None:
            return False
        with self._mutex:
            return len(self._cache) > self._threshold

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __delitem__(self, key):
        return self.delete(key)

    def __contains__(self, key):
        with self._mutex:
            return key in self._cache

    def set_bypassed(self, key):
        with self._mutex:
            self._bypassed.add(key)

    def is_bypassed(self, key):
        with self._mutex:
            return key in self._bypassed


_cache = LessSimpleCache()


def cached(
    timeout=5 * 60,
    key=lambda: "view:%s" % flask.request.path,
    unless=None,
    refreshif=None,
    unless_response=None,
):
    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            logger = logging.getLogger(__name__)

            cache_key = key()

            def f_with_duration(*args, **kwargs):
                start_time = time.time()
                try:
                    return f(*args, **kwargs)
                finally:
                    elapsed = time.time() - start_time
                    logger.debug(
                        "Needed {elapsed:.2f}s to render {path} (key: {key})".format(
                            elapsed=elapsed, path=flask.request.path, key=cache_key
                        )
                    )

            # bypass the cache if "unless" condition is true
            if callable(unless) and unless():
                logger.debug(
                    "Cache for {path} bypassed, calling wrapped function".format(
                        path=flask.request.path
                    )
                )
                _cache.set_bypassed(cache_key)
                return f_with_duration(*args, **kwargs)

            # also bypass the cache if it's disabled completely
            if not settings().getBoolean(["devel", "cache", "enabled"]):
                logger.debug(
                    "Cache for {path} disabled, calling wrapped function".format(
                        path=flask.request.path
                    )
                )
                _cache.set_bypassed(cache_key)
                return f_with_duration(*args, **kwargs)

            rv = _cache.get(cache_key)

            # only take the value from the cache if we are not required to refresh it from the wrapped function
            if rv is not None and (not callable(refreshif) or not refreshif(rv)):
                logger.debug(
                    "Serving entry for {path} from cache (key: {key})".format(
                        path=flask.request.path, key=cache_key
                    )
                )
                if "X-From-Cache" not in rv.headers:
                    rv.headers["X-From-Cache"] = "true"
                return rv

            # get value from wrapped function
            logger.debug(
                "No cache entry or refreshing cache for {path} (key: {key}), calling wrapped function".format(
                    path=flask.request.path, key=cache_key
                )
            )
            rv = f_with_duration(*args, **kwargs)

            # do not store if the "unless_response" condition is true
            if callable(unless_response) and unless_response(rv):
                logger.debug(
                    "Not caching result for {path} (key: {key}), bypassed".format(
                        path=flask.request.path, key=cache_key
                    )
                )
                _cache.set_bypassed(cache_key)
                return rv

            # store it in the cache
            _cache.set(cache_key, rv, timeout=timeout)

            return rv

        return decorated_function

    return decorator


def is_in_cache(key=lambda: "view:%s" % flask.request.path):
    if callable(key):
        key = key()
    return key in _cache


def is_cache_bypassed(key=lambda: "view:%s" % flask.request.path):
    if callable(key):
        key = key()
    return _cache.is_bypassed(key)


def cache_check_headers():
    return "no-cache" in flask.request.cache_control or "no-cache" in flask.request.pragma


def cache_check_response_headers(response):
    if not isinstance(response, flask.Response):
        return False

    headers = response.headers

    if "Cache-Control" in headers and (
        "no-cache" in headers["Cache-Control"] or "no-store" in headers["Cache-Control"]
    ):
        return True

    if "Pragma" in headers and "no-cache" in headers["Pragma"]:
        return True

    if "Expires" in headers and headers["Expires"] in ("0", "-1"):
        return True

    return False


def cache_check_status_code(response, valid):
    if not isinstance(response, flask.Response):
        return False

    if callable(valid):
        return not valid(response.status_code)
    else:
        return response.status_code not in valid


class PreemptiveCache:
    def __init__(self, cachefile):
        self.cachefile = cachefile
        self.environment = None

        self._logger = logging.getLogger(__name__ + "." + self.__class__.__name__)

        self._lock = threading.RLock()

    def record(self, data, unless=None, root=None):
        if callable(unless) and unless():
            return

        entry_data = data
        if callable(entry_data):
            entry_data = entry_data()

        if entry_data is not None:
            if root is None:
                from flask import request

                root = request.path
            self.add_data(root, entry_data)

    def has_record(self, data, root=None):
        if callable(data):
            data = data()

        if data is None:
            return False

        if root is None:
            from flask import request

            root = request.path

        all_data = self.get_data(root)
        for existing in all_data:
            if self._compare_data(data, existing):
                return True

        return False

    def clean_all_data(self, cleanup_function):
        assert callable(cleanup_function)

        with self._lock:
            all_data = self.get_all_data()
            for root, entries in list(all_data.items()):
                old_count = len(entries)
                entries = cleanup_function(root, entries)
                if not entries:
                    del all_data[root]
                    self._logger.debug(f"Removed root {root} from preemptive cache")
                elif len(entries) < old_count:
                    all_data[root] = entries
                    self._logger.debug(
                        "Removed {} entries from preemptive cache for root {}".format(
                            old_count - len(entries), root
                        )
                    )
            self.set_all_data(all_data)

        return all_data

    def get_all_data(self):
        cache_data = None
        with self._lock:
            try:
                cache_data = yaml.load_from_file(path=self.cachefile)
            except OSError as e:
                import errno

                if e.errno != errno.ENOENT:
                    raise
            except Exception:
                self._logger.exception(f"Error while reading {self.cachefile}")

        if cache_data is None:
            cache_data = {}

        if not self._validate_data(cache_data):
            self._logger.warning("Preemptive cache data was invalid, ignoring it")
            cache_data = {}

        return cache_data

    def get_data(self, root):
        cache_data = self.get_all_data()
        return cache_data.get(root, list())

    def set_all_data(self, data):
        from octoprint.util import atomic_write

        with self._lock:
            try:
                with atomic_write(self.cachefile, "wt", max_permissions=0o666) as handle:
                    yaml.save_to_file(data, file=handle, pretty=True)
            except Exception:
                self._logger.exception(f"Error while writing {self.cachefile}")

    def set_data(self, root, data):
        with self._lock:
            all_data = self.get_all_data()
            all_data[root] = data
            self.set_all_data(all_data)

    def add_data(self, root, data):
        def split_matched_and_unmatched(entry, entries):
            matched = []
            unmatched = []

            for e in entries:
                if self._compare_data(e, entry):
                    matched.append(e)
                else:
                    unmatched.append(e)

            return matched, unmatched

        with self._lock:
            cache_data = self.get_all_data()

            if root not in cache_data:
                cache_data[root] = []

            existing, other = split_matched_and_unmatched(data, cache_data[root])

            def get_newest(entries):
                result = None
                for entry in entries:
                    if "_timestamp" in entry and (
                        result is None
                        or (
                            "_timestamp" in result
                            and result["_timestamp"] < entry["_timestamp"]
                        )
                    ):
                        result = entry
                return result

            to_persist = get_newest(existing)
            if not to_persist:
                import copy

                to_persist = copy.deepcopy(data)
                to_persist["_timestamp"] = time.time()
                to_persist["_count"] = 1
                self._logger.info(f"Adding entry for {root} and {to_persist!r}")
            else:
                to_persist["_timestamp"] = time.time()
                to_persist["_count"] = to_persist.get("_count", 0) + 1
                self._logger.debug(
                    f"Updating timestamp and counter for {root} and {data!r}"
                )

            self.set_data(root, [to_persist] + other)

    def _compare_data(self, a, b):
        from octoprint.util import dict_filter

        def strip_ignored(d):
            return dict_filter(d, lambda k, v: not k.startswith("_"))

        return set(strip_ignored(a).items()) == set(strip_ignored(b).items())

    def _validate_data(self, data):
        if not isinstance(data, dict):
            return False

        for entries in data.values():
            if not isinstance(entries, list):
                return False

            for entry in entries:
                if not self._validate_entry(entry):
                    return False

        return True

    def _validate_entry(self, entry):
        return isinstance(entry, dict) and "_timestamp" in entry and "_count" in entry


def preemptively_cached(cache, data, unless=None):
    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            try:
                cache.record(data, unless=unless)
            except Exception:
                logging.getLogger(__name__).exception(
                    f"Error while recording preemptive cache entry: {data!r}"
                )
            return f(*args, **kwargs)

        return decorated_function

    return decorator


def etagged(etag):
    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            rv = f(*args, **kwargs)
            if isinstance(rv, flask.Response):
                try:
                    result = etag
                    if callable(result):
                        result = result(rv)
                    if result:
                        rv.set_etag(result)
                except Exception:
                    logging.getLogger(__name__).exception(
                        "Error while calculating the etag value for response {!r}".format(
                            rv
                        )
                    )
            return rv

        return decorated_function

    return decorator


def lastmodified(date):
    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            rv = f(*args, **kwargs)
            if "Last-Modified" not in rv.headers:
                try:
                    result = date
                    if callable(result):
                        result = result(rv)

                    if not isinstance(result, str):
                        from werkzeug.http import http_date

                        result = http_date(result)

                    if result:
                        rv.headers["Last-Modified"] = result
                except Exception:
                    logging.getLogger(__name__).exception(
                        "Error while calculating the lastmodified value for response {!r}".format(
                            rv
                        )
                    )
            return rv

        return decorated_function

    return decorator


def conditional(condition, met):
    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            try:
                if callable(condition) and condition():
                    # condition has been met, return met-response
                    rv = met
                    if callable(met):
                        rv = met()
                    return rv
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error while evaluating conditional {!r} or met {!r}".format(
                        condition, met
                    )
                )

            # condition hasn't been met, call decorated function
            return f(*args, **kwargs)

        return decorated_function

    return decorator


def with_client_revalidation(f):
    @functools.wraps(f)
    def decorated_function(*args, **kwargs):
        r = f(*args, **kwargs)

        if isinstance(r, flask.Response):
            r = add_revalidation_response_headers(r)

        return r

    return decorated_function


def with_revalidation_checking(
    etag_factory=None, lastmodified_factory=None, condition=None, unless=None
):
    if etag_factory is None:

        def etag_factory(lm=None):
            return None

    if lastmodified_factory is None:

        def lastmodified_factory():
            return None

    if condition is None:

        def condition(lm=None, etag=None):
            if lm is None:
                lm = lastmodified_factory()

            if etag is None:
                etag = etag_factory(lm=lm)

            if flask.request.if_none_match and flask.request.if_modified_since:
                # use both
                return check_lastmodified(lm) and check_etag(etag)
            elif flask.request.if_none_match:
                # use only ETag
                return check_etag(etag)
            elif flask.request.if_modified_since:
                # use only Last-Modified
                return check_lastmodified(lm)
            else:
                # assume stale cache
                return False

    if unless is None:

        def unless():
            return False

    def decorator(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            from octoprint.server import NOT_MODIFIED

            lm = lastmodified_factory()
            etag = etag_factory(lm)

            if condition(lm, etag) and not unless():
                return NOT_MODIFIED

            # generate response
            response = f(*args, **kwargs)

            # set etag header if not already set
            if etag and response.get_etag()[0] is None:
                response.set_etag(etag)

            # set last modified header if not already set
            if lm and response.headers.get("Last-Modified", None) is None:
                if not isinstance(lm, str):
                    from werkzeug.http import http_date

                    lm = http_date(lm)
                response.headers["Last-Modified"] = lm

            response = add_no_max_age_response_headers(response)
            return response

        return decorated_function

    return decorator


def check_etag(etag):
    if etag is None:
        return False

    return (
        flask.request.method in ("GET", "HEAD")
        and flask.request.if_none_match is not None
        and etag in flask.request.if_none_match
    )


def check_lastmodified(lastmodified: Union[int, float, datetime]) -> bool:
    """Compares the provided lastmodified value with the value of the If-Modified-Since header.

    If ``lastmodified`` is an int or float, it's assumed to be a Unix timestamp and converted
    to a timezone aware datetime instance in UTC.

    If ``lastmodified`` is a datetime instance, it needs to be timezone aware or the
    result will always be ``False``.

    Args:
        lastmodified (Union[int, float, datetime]): The last modified value to compare against

    Raises:
        ValueError: If anything but an int, float or datetime instance is passed

    Returns:
        bool: true if the values indicate that the document is still up to date
    """

    if lastmodified is None:
        return False

    if isinstance(lastmodified, (int, float)):
        # max(86400, lastmodified) is workaround for https://bugs.python.org/issue29097,
        # present in CPython 3.6.x up to 3.7.1.
        #
        # I think it's fair to say that we'll never encounter lastmodified values older than
        # 1970-01-02 so this is a safe workaround.
        #
        # Timestamps are defined as seconds since epoch aka 1970/01/01 00:00:00Z, so we
        # use UTC as timezone here.
        lastmodified = datetime.fromtimestamp(
            max(86400, lastmodified), tz=UTC_TZ
        ).replace(microsecond=0)

    if not isinstance(lastmodified, datetime):
        raise ValueError(
            "lastmodified must be a datetime or float or int instance but, got {} instead".format(
                lastmodified.__class__
            )
        )

    if not is_timezone_aware(lastmodified):
        # datetime object is not timezone aware, we can't check lastmodified with that
        logger = logging.getLogger(__name__)
        logger.warning(
            "lastmodified is not timezone aware, cannot check against If-Modified-Since. In the future this will become an error!",
            stack_info=logger.isEnabledFor(logging.DEBUG),
        )
        return False

    return (
        flask.request.method in ("GET", "HEAD")
        and flask.request.if_modified_since is not None
        and lastmodified <= flask.request.if_modified_since
    )


def add_revalidation_response_headers(response):
    import werkzeug.http

    cache_control = werkzeug.http.parse_dict_header(
        response.headers.get("Cache-Control", "")
    )
    if "no-cache" not in cache_control:
        cache_control["no-cache"] = None
    if "must-revalidate" not in cache_control:
        cache_control["must-revalidate"] = None
    response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

    return response


def add_non_caching_response_headers(response):
    import werkzeug.http

    cache_control = werkzeug.http.parse_dict_header(
        response.headers.get("Cache-Control", "")
    )
    if "no-store" not in cache_control:
        cache_control["no-store"] = None
    if "no-cache" not in cache_control:
        cache_control["no-cache"] = None
    if "must-revalidate" not in cache_control:
        cache_control["must-revalidate"] = None
    if "post-check" not in cache_control or cache_control["post-check"] != "0":
        cache_control["post-check"] = "0"
    if "pre-check" not in cache_control or cache_control["pre-check"] != "0":
        cache_control["pre-check"] = "0"
    if "max-age" not in cache_control or cache_control["max-age"] != "0":
        cache_control["max-age"] = "0"
    response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

    response.headers["Pragma"] = "no-cache"
    response.headers["Expires"] = "-1"
    return response


def add_no_max_age_response_headers(response):
    import werkzeug.http

    cache_control = werkzeug.http.parse_dict_header(
        response.headers.get("Cache-Control", "")
    )
    if "max-age" not in cache_control or cache_control["max-age"] != "0":
        cache_control["max-age"] = "0"
    response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

    return response


# ~~ access validators for use with tornado


def permission_validator(request, permission):
    """
    Validates that the given request is made by an authorized user, identified either by API key or existing Flask
    session.

    Must be executed in an existing Flask request context!

    :param request: The Flask request object
    :param request: The required permission
    """

    user = get_flask_user_from_request(request)
    if not user.has_permission(permission):
        raise tornado.web.HTTPError(403)


def permission_and_fresh_credentials_validator(request, permission):
    """
    Validates that the given request is made by an authorized user, identified either by API key or existing Flask
    session, and that the credentials have been checked recently if it's a Flask session.

    Must be executed in an existing Flask request context!

    :param request: The Flask request object
    :param request: The required permission
    """

    permission_validator(request, permission)
    ensure_credentials_checked_recently()


@deprecated(
    "admin_validator is deprecated, please use new permission_validator", since=""
)
def admin_validator(request):
    from octoprint.access.permissions import Permissions

    return permission_validator(request, Permissions.ADMIN)


@deprecated("user_validator is deprecated, please use new permission_validator", since="")
def user_validator(request):
    return True


def get_flask_user_from_request(request):
    """
    Retrieves the current flask user from the request context. Uses API key if available, otherwise the current
    user session if available.

    :param request: flask request from which to retrieve the current user
    :return: the user (might be an anonymous user)
    """
    import flask_login

    import octoprint.server.util

    user = None

    apikey = octoprint.server.util.get_api_key(request)
    if apikey is not None:
        # user from api key?
        user = octoprint.server.util.get_user_for_apikey(apikey)

    if user is None:
        # user still None -> current session user
        user = flask_login.current_user

    if user is None:
        # user still None -> anonymous
        from octoprint.server import userManager

        user = userManager.anonymous_user_factory()

    return user


def redirect_to_tornado(request, target, code=302):
    """
    Redirects from flask to tornado, flask request context must exist.

    :param request:
    :param target:
    :param code:
    :return:
    """

    import flask

    requestUrl = request.url
    appBaseUrl = requestUrl[: requestUrl.find(flask.url_for("index") + "api")]

    redirectUrl = appBaseUrl + target
    if "?" in requestUrl:
        fragment = requestUrl[requestUrl.rfind("?") :]
        redirectUrl += fragment
    return flask.redirect(redirectUrl, code=code)


def restricted_access(func):
    """
    This combines :py:func:`no_firstrun_access` and ``login_required``.
    """

    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        return no_firstrun_access(flask_login.login_required(func))(*args, **kwargs)

    return decorated_view


def no_firstrun_access(func):
    """
    If you decorate a view with this, it will ensure that first setup has been
    done for OctoPrint's Access Control.

    If OctoPrint's Access Control has not been setup yet (indicated by the userManager
    not reporting that its user database has been customized from default), the decorator
    will cause a HTTP 403 status code to be returned by the decorated resource.
    """

    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        # if OctoPrint hasn't been set up yet, abort
        if settings().getBoolean(["server", "firstRun"]) and (
            octoprint.server.userManager is None
            or not octoprint.server.userManager.has_been_customized()
        ):
            flask.abort(403)
        return func(*args, **kwargs)

    return decorated_view


def firstrun_only_access(func):
    """
    If you decorate a view with this, it will ensure that first setup has _not_ been
    done for OctoPrint's Access Control. Otherwise it
    will cause a HTTP 403 status code to be returned by the decorated resource.
    """

    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        # if OctoPrint has been set up yet, abort
        if settings().getBoolean(["server", "firstRun"]) and (
            octoprint.server.userManager is None
            or not octoprint.server.userManager.has_been_customized()
        ):
            return func(*args, **kwargs)
        else:
            flask.abort(403)

    return decorated_view


def credentials_checked_recently():
    minutes = settings().getInt(["accessControl", "defaultReauthenticationTimeout"])
    if not minutes:
        return True

    login_mechanism = flask.session.get("login_mechanism")
    if not octoprint.server.util.LoginMechanism.reauthentication_enabled(login_mechanism):
        return True

    credentials_seen = flask.session.get("credentials_seen")
    now = datetime.now()

    try:
        if credentials_seen and datetime.fromtimestamp(
            credentials_seen
        ) > now - timedelta(minutes=minutes):
            # credentials seen less than the set minutes ago, proceed
            return True
    except Exception:
        logging.getLogger(__name__).exception("Error while checking for seen credentials")
        pass

    return False


def ensure_credentials_checked_recently():
    if not credentials_checked_recently():
        flask.abort(403, description="Please reauthenticate with your credentials")


def require_credentials_checked_recently(func):
    """
    If you decorate a view with this, it will ensure that only users who entered their password
    recently in this login session are allowed to proceed. Otherwise it will cause a HTTP 403 status code
    to be returned by the decorated resource.
    """

    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        ensure_credentials_checked_recently()
        return func(*args, **kwargs)

    return decorated_view


def get_remote_address(request):
    forwardedFor = request.headers.get("X-Forwarded-For", None)
    if forwardedFor is not None:
        return forwardedFor.split(",")[0]
    return request.remote_addr


def get_json_command_from_request(request, valid_commands):
    data = request.get_json()

    if "command" not in data or data["command"] not in valid_commands:
        flask.abort(400, description="command is invalid")

    command = data["command"]
    if any(map(lambda x: x not in data, valid_commands[command])):
        flask.abort(400, description="Mandatory parameters missing")

    return command, data, None


def make_text_response(message, status):
    """
    Helper to generate basic text responses.

    Response will have the provided message as body, the provided status code, and
    a content type of "text/plain".

    Args:
        message: The message in the response body
        status: The HTTP status code

    Returns:

    """
    return make_response(message, status, {"Content-Type": "text/plain"})


def make_api_error(message, status):
    """
    Helper to generate API error responses in JSON format.

    Turns something like ``make_api_error("Not Found", 404)`` into a JSON response
    with body ``{"error": "Not Found"}``.

    Args:
        message: The error message to put into the response
        status: The HTTP status code

    Returns: a flask response to return to the client
    """
    return make_response(flask.jsonify(error=message), status)


##~~ Flask-Assets resolver with plugin asset support


class PluginAssetResolver(flask_assets.FlaskResolver):
    def split_prefix(self, ctx, item):
        app = ctx.environment._app
        if item.startswith("plugin/"):
            try:
                prefix, plugin, name = item.split("/", 2)
                blueprint = prefix + "." + plugin

                directory = flask_assets.get_static_folder(app.blueprints[blueprint])
                item = name
                endpoint = blueprint + ".static"
                return directory, item, endpoint
            except (ValueError, KeyError):
                pass

        return flask_assets.FlaskResolver.split_prefix(self, ctx, item)

    def resolve_output_to_path(self, ctx, target, bundle):
        import os

        return os.path.normpath(os.path.join(ctx.environment.directory, target))


##~~ Webassets updater that takes changes in the configuration into account


class SettingsCheckUpdater(webassets.updater.BaseUpdater):
    updater = "always"

    def __init__(self):
        self._delegate = webassets.updater.get_updater(self.__class__.updater)

    def needs_rebuild(self, bundle, ctx):
        return self._delegate.needs_rebuild(bundle, ctx) or self.changed_settings(ctx)

    def changed_settings(self, ctx):
        if not ctx.cache:
            return False

        cache_key = ("octo", "settings")
        current_hash = settings().effective_hash
        cached_hash = ctx.cache.get(cache_key)
        # This may seem counter-intuitive, but if no cache entry is found
        # then we actually return "no update needed". This is because
        # otherwise if no cache / a dummy cache is used, then we would be
        # rebuilding every single time.
        if cached_hash is not None:
            return cached_hash != current_hash
        return False

    def build_done(self, bundle, ctx):
        self._delegate.build_done(bundle, ctx)
        if not ctx.cache:
            return

        cache_key = ("octo", "settings")
        ctx.cache.set(cache_key, settings().effective_hash)


##~~ core assets collector
def collect_core_assets(preferred_stylesheet="css"):
    assets = {"js": [], "clientjs": [], "css": [], "less": []}
    assets["js"] = [
        "js/app/bindings/allowbindings.js",
        "js/app/bindings/contextmenu.js",
        "js/app/bindings/gettext.js",
        "js/app/bindings/invisible.js",
        "js/app/bindings/popover.js",
        "js/app/bindings/qrcode.js",
        "js/app/bindings/slimscrolledforeach.js",
        "js/app/bindings/toggle.js",
        "js/app/bindings/togglecontent.js",
        "js/app/bindings/valuewithinit.js",
        "js/app/viewmodels/access.js",
        "js/app/viewmodels/appearance.js",
        "js/app/viewmodels/connection.js",
        "js/app/viewmodels/control.js",
        "js/app/viewmodels/files.js",
        "js/app/viewmodels/firstrun_wizard.js",
        "js/app/viewmodels/loginstate.js",
        "js/app/viewmodels/loginui.js",
        "js/app/viewmodels/navigation.js",
        "js/app/viewmodels/printerstate.js",
        "js/app/viewmodels/printerprofiles.js",
        "js/app/viewmodels/settings.js",
        "js/app/viewmodels/slicing.js",
        "js/app/viewmodels/system.js",
        "js/app/viewmodels/temperature.js",
        "js/app/viewmodels/terminal.js",
        "js/app/viewmodels/timelapse.js",
        "js/app/viewmodels/uistate.js",
        "js/app/viewmodels/users.js",
        "js/app/viewmodels/usersettings.js",
        "js/app/viewmodels/wizard.js",
        "js/app/viewmodels/about.js",
    ]

    assets["clientjs"] = [
        "js/app/client/base.js",
        "js/app/client/socket.js",
        "js/app/client/access.js",
        "js/app/client/browser.js",
        "js/app/client/connection.js",
        "js/app/client/control.js",
        "js/app/client/files.js",
        "js/app/client/job.js",
        "js/app/client/languages.js",
        "js/app/client/printer.js",
        "js/app/client/printerprofiles.js",
        "js/app/client/settings.js",
        "js/app/client/slicing.js",
        "js/app/client/system.js",
        "js/app/client/timelapse.js",
        "js/app/client/users.js",
        "js/app/client/util.js",
        "js/app/client/wizard.js",
    ]

    if preferred_stylesheet == "less":
        assets["less"].append("less/octoprint.less")
    elif preferred_stylesheet == "css":
        assets["css"].append("css/octoprint.css")

    return assets


##~~ plugin assets collector


def collect_plugin_assets(preferred_stylesheet="css"):
    logger = logging.getLogger(__name__ + ".collect_plugin_assets")

    supported_stylesheets = ("css", "less")
    assets = {
        "bundled": {
            "js": DefaultOrderedDict(list),
            "clientjs": DefaultOrderedDict(list),
            "css": DefaultOrderedDict(list),
            "less": DefaultOrderedDict(list),
        },
        "external": {
            "js": DefaultOrderedDict(list),
            "clientjs": DefaultOrderedDict(list),
            "css": DefaultOrderedDict(list),
            "less": DefaultOrderedDict(list),
        },
    }

    asset_plugins = octoprint.plugin.plugin_manager().get_implementations(
        octoprint.plugin.AssetPlugin
    )
    for implementation in asset_plugins:
        name = implementation._identifier
        is_bundled = implementation._plugin_info.bundled

        asset_key = "bundled" if is_bundled else "external"

        try:
            all_assets = implementation.get_assets()
            basefolder = implementation.get_asset_folder()
        except Exception:
            logger.exception(
                "Got an error while trying to collect assets from {}, ignoring assets from the plugin".format(
                    name
                ),
                extra={"plugin": name},
            )
            continue

        def asset_exists(category, asset):
            exists = os.path.exists(os.path.join(basefolder, asset))
            if not exists:
                logger.warning(
                    "Plugin {} is referring to non existing {} asset {}".format(
                        name, category, asset
                    )
                )
            return exists

        if "js" in all_assets:
            for asset in all_assets["js"]:
                if not asset_exists("js", asset):
                    continue
                assets[asset_key]["js"][name].append(f"plugin/{name}/{asset}")

        if "clientjs" in all_assets:
            for asset in all_assets["clientjs"]:
                if not asset_exists("clientjs", asset):
                    continue
                assets[asset_key]["clientjs"][name].append(f"plugin/{name}/{asset}")

        if preferred_stylesheet in all_assets:
            for asset in all_assets[preferred_stylesheet]:
                if not asset_exists(preferred_stylesheet, asset):
                    continue
                assets[asset_key][preferred_stylesheet][name].append(
                    f"plugin/{name}/{asset}"
                )
        else:
            for stylesheet in supported_stylesheets:
                if stylesheet not in all_assets:
                    continue

                for asset in all_assets[stylesheet]:
                    if not asset_exists(stylesheet, asset):
                        continue
                    assets[asset_key][stylesheet][name].append(f"plugin/{name}/{asset}")
                break

    return assets


##~~ JSON encoding


class OctoPrintJsonProvider(flask.json.provider.DefaultJSONProvider):
    @staticmethod
    def default(object_):
        try:
            return JsonEncoding.encode(object_)
        except TypeError:
            return flask.json.provider.DefaultJSONProvider.default(object_)

    def dumps(self, obj: Any, **kwargs: Any) -> str:
        kwargs.setdefault("allow_nan", False)
        return super().dumps(obj, **kwargs)


##~~ Session signing


def session_signature(user, session):
    from octoprint.server import userManager

    key = userManager.signature_key_for_user(user, current_app.config["SECRET_KEY"])
    return hmac.new(
        key.encode("utf-8"), session.encode("utf-8"), hashlib.sha512
    ).hexdigest()


def validate_session_signature(sig, user, session):
    user_sig = session_signature(user, session)
    return len(user_sig) == len(sig) and hmac.compare_digest(sig, user_sig)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.sessioncontainer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Simple heapq-based session implementation with sliding expiration window
    support.
"""

from heapq import heappush, heappop
from time import time
from hashlib import md5
from random import random


def _random_key():
    """Return random session key"""
    i = md5()
    i.update('%s%s' % (random(), time()))
    return i.hexdigest()


class SessionMixin(object):
    """Represents one session object stored in the session container.
    Derive from this object to store additional data.
    """

    def __init__(self, session_id=None, expiry=None):
        """Constructor.

        ``session_id``
            Optional session id. If not provided, will generate
            new session id.
        ``expiry``
            Expiration time. If not provided, will never expire.
        """
        self.session_id = session_id or _random_key()
        self.promoted = None
        self.expiry = expiry

        if self.expiry is not None:
            self.expiry_date = time() + self.expiry

    def is_alive(self):
        """Check if session is still alive"""
        return self.expiry_date > time()

    def promote(self):
        """Mark object as alive, so it won't be collected during next
        run of the garbage collector.
        """
        if self.expiry is not None:
            self.promoted = time() + self.expiry

    def on_delete(self, forced):
        """Triggered when object was expired or deleted."""
        pass

    def __lt__(self, other):
        return self.expiry_date < other.expiry_date
    
    __cmp__ =  __lt__

    def __repr__(self):
        return '%f %s %d' % (getattr(self, 'expiry_date', -1),
                             self.session_id,
                             self.promoted or 0)


class SessionContainer(object):
    """Session container object.

    If we will implement sessions with Tornado timeouts, for polling transports
    it will be nightmare - if load will be high, number of discarded timeouts
    will be huge and will be huge performance hit, as Tornado will have to
    clean them up all the time.
    """
    def __init__(self):
        self._items = {}
        self._queue = []

    def add(self, session):
        """Add session to the container.

        `session`
            Session object
        """
        self._items[session.session_id] = session

        if session.expiry is not None:
            heappush(self._queue, session)

    def get(self, session_id):
        """Return session object or None if it is not available

        `session_id`
            Session identifier
        """
        return self._items.get(session_id, None)

    def remove(self, session_id):
        """Remove session object from the container

        `session_id`
            Session identifier
        """
        session = self._items.get(session_id, None)

        if session is not None:
            session.promoted = -1
            session.on_delete(True)
            del self._items[session_id]
            return True

        return False

    def expire(self, current_time=None):
        """Expire any old entries

        `current_time`
            Optional time to be used to clean up queue (can be used in unit tests)
        """
        if not self._queue:
            return

        if current_time is None:
            current_time = time()

        while self._queue:
            # Get top most item
            top = self._queue[0]

            # Early exit if item was not promoted and its expiration time
            # is greater than now.
            if top.promoted is None and top.expiry_date > current_time:
                break

            # Pop item from the stack
            top = heappop(self._queue)

            need_reschedule = (top.promoted is not None
                               and top.promoted > current_time)

            # Give chance to reschedule
            if not need_reschedule:
                top.promoted = None
                top.on_delete(False)

                need_reschedule = (top.promoted is not None
                                   and top.promoted > current_time)

            # If item is promoted and expiration time somewhere in future
            # just reschedule it
            if need_reschedule:
                top.expiry_date = top.promoted
                top.promoted = None
                heappush(self._queue, top)
            else:
                del self._items[top.session_id]



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import Dict, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class SlicingConfig(BaseModel):
    enabled: bool = True
    """Whether to enable slicing support or not."""

    defaultSlicer: Optional[str] = None
    """Default slicer to use."""

    defaultProfiles: Dict[str, str] = {}
    """Default slicing profiles per slicer, maps slicer identifier to profile identifier."""



"""
This module defines the interface for communicating with a connected printer.

The communication is in fact divided in two components, the :class:`PrinterInterface` and a deeper lying
communication layer. However, plugins should only ever need to use the :class:`PrinterInterface` as the
abstracted version of the actual printer communication.

.. autofunction:: get_connection_options

.. autoclass:: PrinterInterface
   :members:

.. autoclass:: PrinterCallback
   :members:
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import re

from octoprint.filemanager import FileDestinations
from octoprint.settings import settings
from octoprint.util import deprecated, natural_key


@deprecated(
    message="get_connection_options has been replaced by PrinterInterface.get_connection_options",
    includedoc="Replaced by :func:`PrinterInterface.get_connection_options`",
    since="1.3.0",
)
def get_connection_options():
    return PrinterInterface.get_connection_options()


class PrinterInterface:
    """
    The :class:`PrinterInterface` represents the developer interface to the :class:`~octoprint.printer.standard.Printer`
    instance.
    """

    valid_axes = ("x", "y", "z", "e")
    """Valid axes identifiers."""

    valid_tool_regex = re.compile(r"^(tool\d+)$")
    """Regex for valid tool identifiers."""

    valid_heater_regex = re.compile(r"^(tool\d+|bed|chamber)$")
    """Regex for valid heater identifiers."""

    @classmethod
    def get_connection_options(cls, *args, **kwargs):
        """
        Retrieves the available ports, baudrates, preferred port and baudrate for connecting to the printer.

        Returned ``dict`` has the following structure::

            ports: <list of available serial ports>
            baudrates: <list of available baudrates>
            portPreference: <configured default serial port>
            baudratePreference: <configured default baudrate>
            autoconnect: <whether autoconnect upon server startup is enabled or not>

        Returns:
            (dict): A dictionary holding the connection options in the structure specified above
        """
        import octoprint.util.comm as comm

        return {
            "ports": sorted(comm.serialList(), key=natural_key),
            "baudrates": sorted(comm.baudrateList(), reverse=True),
            "portPreference": settings().get(["serial", "port"]),
            "baudratePreference": settings().getInt(["serial", "baudrate"]),
            "autoconnect": settings().getBoolean(["serial", "autoconnect"]),
        }

    def connect(self, port=None, baudrate=None, profile=None, *args, **kwargs):
        """
        Connects to the printer, using the specified serial ``port``, ``baudrate`` and printer ``profile``. If a
        connection is already established, that connection will be closed prior to connecting anew with the provided
        parameters.

        Arguments:
            port (str): Name of the serial port to connect to. If not provided, an auto detection will be attempted.
            baudrate (int): Baudrate to connect with. If not provided, an auto detection will be attempted.
            profile (str): Name of the printer profile to use for this connection. If not provided, the default
                will be retrieved from the :class:`PrinterProfileManager`.
        """
        pass

    def disconnect(self, *args, **kwargs):
        """
        Disconnects from the printer. Does nothing if no connection is currently established.
        """
        raise NotImplementedError()

    def get_transport(self, *args, **kwargs):
        """
        Returns the communication layer's transport object, if a connection is currently established.

        Note that this doesn't have to necessarily be a :class:`serial.Serial` instance, it might also be something
        different, so take care to do instance checks before attempting to access any properties or methods.

        Returns:
            object: The communication layer's transport object
        """
        raise NotImplementedError()

    def job_on_hold(self, blocking=True, *args, **kwargs):
        """
        Contextmanager that allows executing code while printing while making sure that no commands from the file
        being printed are continued to be sent to the printer. Note that this will only work for local files,
        NOT SD files.

        Example:

        .. code-block:: python

           with printer.job_on_hold():
               park_printhead()
               take_snapshot()
               send_printhead_back()

        It should be used sparingly and only for very specific situations (such as parking the print head somewhere,
        taking a snapshot from the webcam, then continuing). If you abuse this, you WILL cause print quality issues!

        A lock is in place that ensures that the context can only actually be held by one thread at a time. If you
        don't want to block on acquire, be sure to set ``blocking`` to ``False`` and catch the ``RuntimeException`` thrown
        if the lock can't be acquired.

        Args:
                blocking (bool): Whether to block while attempting to acquire the lock (default) or not
        """
        raise NotImplementedError()

    def set_job_on_hold(self, value, blocking=True, *args, **kwargs):
        """
        Setter for finer control over putting jobs on hold. Set to ``True`` to ensure that no commands from the file
        being printed are continued to be sent to the printer. Set to ``False`` to resume. Note that this will only
        work for local files, NOT SD files.

        Make absolutely sure that if you set this flag, you will always also unset it again. If you don't, the job will
        be stuck forever.

        Example:

        .. code-block:: python

           if printer.set_job_on_hold(True):
               try:
                   park_printhead()
                   take_snapshot()
                   send_printhead_back()
               finally:
                   printer.set_job_on_hold(False)

        Just like :func:`~octoprint.printer.PrinterInterface.job_on_hold` this should be used sparingly and only for
        very specific situations. If you abuse this, you WILL cause print quality issues!

        Args:
                value (bool): The value to set
                blocking (bool): Whether to block while attempting to set the value (default) or not

        Returns:
                (bool) Whether the value could be set successfully (True) or a timeout was encountered (False)
        """
        raise NotImplementedError()

    def fake_ack(self, *args, **kwargs):
        """
        Fakes an acknowledgment for the communication layer. If the communication between OctoPrint and the printer
        gets stuck due to lost "ok" responses from the server due to communication issues, this can be used to get
        things going again.
        """
        raise NotImplementedError()

    def commands(self, commands, tags=None, force=False, *args, **kwargs):
        """
        Sends the provided ``commands`` to the printer.

        Arguments:
            commands (str, list): The commands to send. Might be a single command provided just as a string or a list
                of multiple commands to send in order.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
            force (bool): Whether to force sending of the command right away or allow queuing while printing
        """
        raise NotImplementedError()

    def script(self, name, context=None, tags=None, *args, **kwargs):
        """
        Sends the GCODE script ``name`` to the printer.

        The script will be run through the template engine, the rendering context can be extended by providing a
        ``context`` with additional template variables to use.

        If the script is unknown, an :class:`UnknownScriptException` will be raised.

        Arguments:
            name (str): The name of the GCODE script to render.
            context (dict): An optional context of additional template variables to provide to the renderer.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle

        Raises:
            UnknownScriptException: There is no GCODE script with name ``name``
        """
        raise NotImplementedError()

    def jog(self, axes, relative=True, speed=None, tags=None, *args, **kwargs):
        """
        Jogs the specified printer ``axis`` by the specified ``amount`` in mm.

        Arguments:
            axes (dict): Axes and distances to jog, keys are axes ("x", "y", "z"), values are distances in mm
            relative (bool): Whether to interpret the distance values as relative (true, default) or absolute (false)
                coordinates
            speed (int, bool or None): Speed at which to jog (F parameter). If set to ``False`` no speed will be set
                specifically. If set to ``None`` (or left out) the minimum of all involved axes speeds from the printer
                profile will be used.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def home(self, axes, tags=None, *args, **kwargs):
        """
        Homes the specified printer ``axes``.

        Arguments:
            axes (str, list): The axis or axes to home, each of which must converted to lower case must match one of
                "x", "y", "z" and "e"
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def extrude(self, amount, speed=None, tags=None, *args, **kwargs):
        """
        Extrude ``amount`` millimeters of material from the tool.

        Arguments:
            amount (int, float): The amount of material to extrude in mm
            speed (int, None): Speed at which to extrude (F parameter). If set to ``None`` (or left out)
            the maximum speed of E axis from the printer profile will be used.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def change_tool(self, tool, tags=None, *args, **kwargs):
        """
        Switch the currently active ``tool`` (for which extrude commands will apply).

        Arguments:
            tool (str): The tool to switch to, matching the regex "tool[0-9]+" (e.g. "tool0", "tool1", ...)
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def set_temperature(self, heater, value, tags=None, *args, **kwargs):
        """
        Sets the target temperature on the specified ``heater`` to the given ``value`` in celsius.

        Arguments:
            heater (str): The heater for which to set the target temperature. Either "bed" for setting the bed
                temperature, "chamber" for setting the temperature of the heated enclosure or something matching the
                regular expression "tool[0-9]+" (e.g. "tool0", "tool1", ...) for the hotends of the printer. However,
                addressing components that are disabled or unconfigured in the printer profile will result in a
                "Suppressed command" error popup message.
            value (int, float): The temperature in celsius to set the target temperature to.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle.
        """
        raise NotImplementedError()

    def set_temperature_offset(self, offsets=None, tags=None, *args, **kwargs):
        """
        Sets the temperature ``offsets`` to apply to target temperatures read from a GCODE file while printing.

        Arguments:
            offsets (dict): A dictionary specifying the offsets to apply. Keys must match the format for the ``heater``
                parameter to :func:`set_temperature`, so "bed" for the offset for the bed target temperature and
                "tool[0-9]+" for the offsets to the hotend target temperatures.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def feed_rate(self, factor, tags=None, *args, **kwargs):
        """
        Sets the ``factor`` for the printer's feed rate.

        Arguments:
            factor (int, float): The factor for the feed rate to send to the firmware. Percentage expressed as either an
                int between 0 and 100 or a float between 0 and 1.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def flow_rate(self, factor, tags=None, *args, **kwargs):
        """
        Sets the ``factor`` for the printer's flow rate.

        Arguments:
            factor (int, float): The factor for the flow rate to send to the firmware. Percentage expressed as either an
                int between 0 and 100 or a float between 0 and 1.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def can_modify_file(self, path, sd, *args, **kwargs):
        """
        Determines whether the ``path`` (on the printer's SD if ``sd`` is True) may be modified (updated or deleted)
        or not.

        A file that is currently being printed is not allowed to be modified. Any other file or the current file
        when it is not being printed is fine though.

        :since: 1.3.2

        .. warning::

           This was introduced in 1.3.2 to work around an issue when updating a file that is already selected.
           I'm not 100% sure at this point if this is the best approach to solve this issue, so if you decide
           to depend on this particular method in this interface, be advised that it might vanish in future
           versions!

        Arguments:
            path (str): path in storage of the file to check
            sd (bool): True if to check against SD storage, False otherwise

        Returns:
            (bool) True if the file may be modified, False otherwise
        """
        return not (
            self.is_current_file(path, sd) and (self.is_printing() or self.is_paused())
        )

    def is_current_file(self, path, sd, *args, **kwargs):
        """
        Returns whether the provided ``path`` (on the printer's SD if ``sd`` is True) is the currently selected
        file for printing.

        :since: 1.3.2

        .. warning::

           This was introduced in 1.3.2 to work around an issue when updating a file that is already selected.
           I'm not 100% sure at this point if this is the best approach to solve this issue, so if you decide
           to depend on this particular method in this interface, be advised that it might vanish in future
           versions!

        Arguments:
            path (str): path in storage of the file to check
            sd (bool): True if to check against SD storage, False otherwise

        Returns:
            (bool) True if the file is currently selected, False otherwise
        """
        current_job = self.get_current_job()
        if current_job is not None and "file" in current_job:
            current_job_file = current_job["file"]
            if "path" in current_job_file and "origin" in current_job_file:
                current_file_path = current_job_file["path"]
                current_file_origin = current_job_file["origin"]

                return path == current_file_path and sd == (
                    current_file_origin == FileDestinations.SDCARD
                )

        return False

    def select_file(
        self, path, sd, printAfterSelect=False, pos=None, tags=None, *args, **kwargs
    ):
        """
        Selects the specified ``path`` for printing, specifying if the file is to be found on the ``sd`` or not.
        Optionally can also directly start the print after selecting the file.

        Arguments:
            path (str): The path to select for printing. Either an absolute path or relative path to a  local file in
                the uploads folder or a filename on the printer's SD card.
            sd (boolean): Indicates whether the file is on the printer's SD card or not.
            printAfterSelect (boolean): Indicates whether a print should be started
                after the file is selected.
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle

        Raises:
            InvalidFileType: if the file is not a machinecode file and hence cannot be printed
            InvalidFileLocation: if an absolute path was provided and not contained within local storage or
                doesn't exist
        """
        raise NotImplementedError()

    def unselect_file(self, *args, **kwargs):
        """
        Unselects and currently selected file.
        """
        raise NotImplementedError()

    def start_print(self, tags=None, *args, **kwargs):
        """
        Starts printing the currently selected file. If no file is currently selected, does nothing.

        Arguments:
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def pause_print(self, tags=None, *args, **kwargs):
        """
        Pauses the current print job if it is currently running, does nothing otherwise.

        Arguments:
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def resume_print(self, tags=None, *args, **kwargs):
        """
        Resumes the current print job if it is currently paused, does nothing otherwise.

        Arguments:
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def toggle_pause_print(self, tags=None, *args, **kwargs):
        """
        Pauses the current print job if it is currently running or resumes it if it is currently paused.

        Arguments:
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        if self.is_printing():
            self.pause_print(tags=tags, *args, **kwargs)
        elif self.is_paused():
            self.resume_print(tags=tags, *args, **kwargs)

    def cancel_print(self, tags=None, *args, **kwargs):
        """
        Cancels the current print job.

        Arguments:
            tags (set of str): An optional set of tags to attach to the command(s) throughout their lifecycle
        """
        raise NotImplementedError()

    def log_lines(self, *lines):
        """
        Logs the provided lines to the printer log and serial.log
        Args:
                *lines: the lines to log
        """
        pass

    def get_state_string(self, *args, **kwargs):
        """
        Returns:
             (str) A human readable string corresponding to the current communication state.
        """
        raise NotImplementedError()

    def get_state_id(self, *args, **kwargs):
        """
        Identifier of the current communication state.

        Possible values are:

          * OPEN_SERIAL
          * DETECT_SERIAL
          * DETECT_BAUDRATE
          * CONNECTING
          * OPERATIONAL
          * PRINTING
          * PAUSED
          * CLOSED
          * ERROR
          * CLOSED_WITH_ERROR
          * TRANSFERING_FILE
          * OFFLINE
          * UNKNOWN
          * NONE

        Returns:
             (str) A unique identifier corresponding to the current communication state.
        """

    def get_current_data(self, *args, **kwargs):
        """
        Returns:
            (dict) The current state data.
        """
        raise NotImplementedError()

    def get_current_job(self, *args, **kwargs):
        """
        Returns:
            (dict) The data of the current job.
        """
        raise NotImplementedError()

    def get_current_temperatures(self, *args, **kwargs):
        """
        Returns:
            (dict) The current temperatures.
        """
        raise NotImplementedError()

    def get_temperature_history(self, *args, **kwargs):
        """
        Returns:
            (list) The temperature history.
        """
        raise NotImplementedError()

    def get_current_connection(self, *args, **kwargs):
        """
        Returns:
            (tuple) The current connection information as a 4-tuple ``(connection_string, port, baudrate, printer_profile)``.
                If the printer is currently not connected, the tuple will be ``("Closed", None, None, None)``.
        """
        raise NotImplementedError()

    def is_closed_or_error(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently disconnected and/or in an error state.
        """
        raise NotImplementedError()

    def is_operational(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently connected and available.
        """
        raise NotImplementedError()

    def is_printing(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently printing.
        """
        raise NotImplementedError()

    def is_cancelling(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently cancelling a print.
        """
        raise NotImplementedError()

    def is_pausing(self, *args, **kwargs):
        """
        Returns:
                (boolean) Whether the printer is currently pausing a print.
        """
        raise NotImplementedError()

    def is_paused(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently paused.
        """
        raise NotImplementedError()

    def is_error(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently in an error state.
        """
        raise NotImplementedError()

    def is_ready(self, *args, **kwargs):
        """
        Returns:
            (boolean) Whether the printer is currently operational and ready for new print jobs (not printing).
        """
        raise NotImplementedError()

    def register_callback(self, callback, *args, **kwargs):
        """
        Registers a :class:`PrinterCallback` with the instance.

        Arguments:
            callback (PrinterCallback): The callback object to register.
        """
        raise NotImplementedError()

    def unregister_callback(self, callback, *args, **kwargs):
        """
        Unregisters a :class:`PrinterCallback` from the instance.

        Arguments:
            callback (PrinterCallback): The callback object to unregister.
        """
        raise NotImplementedError()

    def send_initial_callback(self, callback):
        """
        Sends the initial printer update to :class:`PrinterCallback`.

        Arguments:
                callback (PrinterCallback): The callback object to send initial data to.
        """
        raise NotImplementedError()

    @property
    def firmware_info(self):
        raise NotImplementedError()

    @property
    def error_info(self):
        raise NotImplementedError()


class PrinterCallback:
    def on_printer_add_log(self, data):
        """
        Called when the :class:`PrinterInterface` receives a new communication log entry from the communication layer.

        Arguments:
            data (str): The received log line.
        """
        pass

    def on_printer_add_message(self, data):
        """
        Called when the :class:`PrinterInterface` receives a new message from the communication layer.

        Arguments:
            data (str): The received message.
        """
        pass

    def on_printer_add_temperature(self, data):
        """
        Called when the :class:`PrinterInterface` receives a new temperature data set from the communication layer.

        ``data`` is a ``dict`` of the following structure::

            tool0:
                actual: <temperature of the first hotend, in degC>
                target: <target temperature of the first hotend, in degC>
            ...
            bed:
                actual: <temperature of the bed, in degC>
                target: <target temperature of the bed, in degC>
            chamber:
                actual: <temperature of the chamber, in degC>
                target: <target temperature of the chamber, in degC>

        Arguments:
            data (dict): A dict of all current temperatures in the format as specified above
        """
        pass

    def on_printer_received_registered_message(self, name, output):
        """
        Called when the :class:`PrinterInterface` received a registered message, e.g. from a feedback command.

        Arguments:
            name (str): Name of the registered message (e.g. the feedback command)
            output (str): Output for the registered message
        """
        pass

    def on_printer_send_initial_data(self, data):
        """
        Called when registering as a callback with the :class:`PrinterInterface` to receive the initial data (state,
        log and temperature history etc) from the printer.

        ``data`` is a ``dict`` of the following structure::

            temps:
              - time: <timestamp of the temperature data point>
                tool0:
                    actual: <temperature of the first hotend, in degC>
                    target: <target temperature of the first hotend, in degC>
                ...
                bed:
                    actual: <temperature of the bed, in degC>
                    target: <target temperature of the bed, in degC>
              - ...
            logs: <list of current communication log lines>
            messages: <list of current messages from the firmware>

        Arguments:
            data (dict): The initial data in the format as specified above.
        """
        pass

    def on_printer_send_current_data(self, data):
        """
        Called when the internal state of the :class:`PrinterInterface` changes, due to changes in the printer state,
        temperatures, log lines, job progress etc. Updates via this method are guaranteed to be throttled to a maximum
        of 2 calls per second.

        ``data`` is a ``dict`` of the following structure::

            state:
                text: <current state string>
                flags:
                    operational: <whether the printer is currently connected and responding>
                    printing: <whether the printer is currently printing>
                    closedOrError: <whether the printer is currently disconnected and/or in an error state>
                    error: <whether the printer is currently in an error state>
                    paused: <whether the printer is currently paused>
                    ready: <whether the printer is operational and ready for jobs>
                    sdReady: <whether an SD card is present>
            job:
                file:
                    name: <name of the file>,
                    size: <size of the file in bytes>,
                    origin: <origin of the file, "local" or "sdcard">,
                    date: <last modification date of the file>
                estimatedPrintTime: <estimated print time of the file in seconds>
                lastPrintTime: <last print time of the file in seconds>
                filament:
                    length: <estimated length of filament needed for this file, in mm>
                    volume: <estimated volume of filament needed for this file, in ccm>
            progress:
                completion: <progress of the print job in percent (0-100)>
                filepos: <current position in the file in bytes>
                printTime: <current time elapsed for printing, in seconds>
                printTimeLeft: <estimated time left to finish printing, in seconds>
            currentZ: <current position of the z axis, in mm>
            offsets: <current configured temperature offsets, keys are "bed" or "tool[0-9]+", values the offset in degC>

        Arguments:
            data (dict): The current data in the format as specified above.
        """
        pass


class UnknownScript(Exception):
    def __init__(self, name, *args, **kwargs):
        self.name = name


class InvalidFileLocation(Exception):
    pass


class InvalidFileType(Exception):
    pass



import os
import threading
import time
from collections import defaultdict

import flask
from flask_babel import gettext

import octoprint.plugin
from octoprint.access import ADMIN_GROUP, USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.server import NO_CONTENT, current_user
from octoprint.server.util import require_fresh_login_with
from octoprint.server.util.flask import (
    add_non_caching_response_headers,
    credentials_checked_recently,
    ensure_credentials_checked_recently,
    no_firstrun_access,
    restricted_access,
)
from octoprint.settings import valid_boolean_trues
from octoprint.util import ResettableTimer, atomic_write, generate_api_key, yaml

CUTOFF_TIME = 10 * 60  # 10min
POLL_TIMEOUT = 5  # 5 seconds


class AppAlreadyExists(Exception):
    pass


class PendingDecision:
    def __init__(self, app_id, app_token, user_id, user_token, timeout_callback=None):
        self.app_id = app_id
        self.app_token = app_token
        self.user_id = user_id
        self.user_token = user_token
        self.created = time.monotonic()

        if callable(timeout_callback):
            self.poll_timeout = ResettableTimer(
                POLL_TIMEOUT, timeout_callback, [user_token]
            )
            self.poll_timeout.start()

    def external(self):
        return {
            "app_id": self.app_id,
            "user_id": self.user_id,
            "user_token": self.user_token,
        }

    def __repr__(self):
        return "PendingDecision({!r}, {!r}, {!r}, {!r}, timeout_callback=...)".format(
            self.app_id, self.app_token, self.user_id, self.user_token
        )


class ReadyDecision:
    def __init__(self, app_id, app_token, user_id):
        self.app_id = app_id
        self.app_token = app_token
        self.user_id = user_id

    @classmethod
    def for_pending(cls, pending, user_id):
        return cls(pending.app_id, pending.app_token, user_id)

    def __repr__(self):
        return "ReadyDecision({!r}, {!r}, {!r})".format(
            self.app_id, self.app_token, self.user_id
        )


class ActiveKey:
    def __init__(self, app_id, api_key, user_id):
        self.app_id = app_id
        self.api_key = api_key
        self.user_id = user_id

    def external(self, incl_key=False):
        result = {"app_id": self.app_id, "user_id": self.user_id}
        if incl_key:
            result["api_key"] = self.api_key
        return result

    def internal(self):
        return {"app_id": self.app_id, "api_key": self.api_key}

    @classmethod
    def for_internal(cls, internal, user_id):
        return cls(internal["app_id"], internal["api_key"], user_id)

    def __repr__(self):
        return "ActiveKey({!r}, {!r}, {!r})".format(
            self.app_id, self.api_key, self.user_id
        )


class AppKeysPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.TemplatePlugin,
):
    def __init__(self):
        self._pending_decisions = []
        self._pending_lock = threading.RLock()

        self._ready_decisions = []
        self._ready_lock = threading.RLock()

        self._keys = defaultdict(list)
        self._keys_lock = threading.RLock()

        self._key_path = None

    def initialize(self):
        self._key_path = os.path.join(self.get_plugin_data_folder(), "keys.yaml")
        self._load_keys()

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "ADMIN",
                "name": "Admin access",
                "description": gettext("Allows administrating all application keys"),
                "roles": ["admin"],
                "dangerous": True,
                "default_groups": [ADMIN_GROUP],
            },
            {
                "key": "GRANT",
                "name": "Grant access",
                "description": gettext("Allows to grant app access"),
                "roles": ["user"],
                "default_groups": [USER_GROUP],
            },
        ]

    ##~~ TemplatePlugin

    def get_template_configs(self):
        return [
            {"type": "usersettings", "name": gettext("Application Keys")},
            {"type": "settings", "name": gettext("Application Keys")},
        ]

    ##~~ AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/appkeys.js"],
            "clientjs": ["clientjs/appkeys.js"],
            "less": ["less/appkeys.less"],
            "css": ["css/appkeys.css"],
        }

    ##~~ BlueprintPlugin mixin

    @octoprint.plugin.BlueprintPlugin.route("/probe", methods=["GET"])
    @no_firstrun_access
    def handle_probe(self):
        return NO_CONTENT

    @octoprint.plugin.BlueprintPlugin.route("/request", methods=["POST"])
    @octoprint.plugin.BlueprintPlugin.csrf_exempt()
    @no_firstrun_access
    def handle_request(self):
        data = flask.request.json
        if data is None:
            flask.abort(400, description="Missing key request")

        if "app" not in data:
            flask.abort(400, description="No app name provided")

        app_name = data["app"]
        user_id = None
        if "user" in data and data["user"]:
            user_id = data["user"]

        app_token, user_token = self._add_pending_decision(app_name, user_id=user_id)
        auth_dialog = flask.url_for(
            "plugin.appkeys.handle_auth_dialog", app_token=app_token, _external=True
        ) + (f"?user={user_id}" if user_id else "")

        self._plugin_manager.send_plugin_message(
            self._identifier,
            {
                "type": "request_access",
                "app_name": app_name,
                "user_token": user_token,
                "user_id": user_id,
            },
        )
        response = flask.jsonify(app_token=app_token, auth_dialog=auth_dialog)
        response.status_code = 201
        response.headers["Location"] = flask.url_for(
            ".handle_decision_poll", app_token=app_token, _external=True
        )
        return response

    @octoprint.plugin.BlueprintPlugin.route("/request/<app_token>", methods=["GET"])
    @no_firstrun_access
    def handle_decision_poll(self, app_token):
        result = self._get_pending_by_app_token(app_token)
        if result:
            for pending_decision in result:
                pending_decision.poll_timeout.reset()

            response = flask.jsonify(message="Awaiting decision")
            response.status_code = 202
            return response

        result = self._get_decision(app_token)
        if result:
            return flask.jsonify(api_key=result)

        return flask.abort(404)

    @octoprint.plugin.BlueprintPlugin.route("/auth/<app_token>", methods=["GET"])
    @no_firstrun_access
    def handle_auth_dialog(self, app_token):
        from octoprint.server.util.csrf import add_csrf_cookie

        user_id = current_user.get_name()
        required_user = flask.request.args.get("user", None)

        pendings = self._get_pending_by_app_token(app_token)
        if not pendings:
            return flask.abort(404)

        response = require_fresh_login_with(
            permissions=[Permissions.PLUGIN_APPKEYS_GRANT], user_id=required_user
        )
        if response:
            return response

        pending = None
        for p in pendings:
            if p.user_id == required_user or (not required_user and p.user_id == user_id):
                pending = p
                break
        else:
            return flask.abort(404)

        app_id = pending.app_id
        user_token = pending.user_token
        redirect_url = flask.request.args.get("redirect", "")

        response = flask.make_response(
            flask.render_template(
                "plugin_appkeys/appkeys_authdialog.jinja2",
                app=app_id,
                user=user_id,
                user_token=user_token,
                redirect_url=redirect_url,
                theming=[],
                request_text=gettext(
                    '"<strong>%(app)s</strong>" has requested access to control OctoPrint through the API.'
                ).replace("%(app)s", app_id),
            )
        )
        return add_csrf_cookie(add_non_caching_response_headers(response))

    @octoprint.plugin.BlueprintPlugin.route("/decision/<user_token>", methods=["POST"])
    @restricted_access
    def handle_decision(self, user_token):
        data = flask.request.json
        if "decision" not in data:
            flask.abort(400, description="No decision provided")

        if not Permissions.PLUGIN_APPKEYS_GRANT.can():
            flask.abort(403, description="No permission to grant app access")

        ensure_credentials_checked_recently()

        decision = data["decision"] in valid_boolean_trues
        user_id = current_user.get_name()

        result = self._set_decision(user_token, decision, user_id)
        if not result:
            return flask.abort(404)

        # Close access_request dialog for this request on all open OctoPrint connections
        self._plugin_manager.send_plugin_message(
            self._identifier, {"type": "end_request", "user_token": user_token}
        )

        return NO_CONTENT

    def is_blueprint_protected(self):
        return False  # No API key required to request API access

    def is_blueprint_csrf_protected(self):
        return True  # protect anything that isn't explicitly marked as exempt

    ##~~ SimpleApiPlugin mixin

    def get_api_commands(self):
        return {"generate": ["app"], "revoke": []}

    def on_api_get(self, request):
        user_id = current_user.get_name()
        if not user_id:
            return flask.abort(403)

        # GET ?app_id=...[&user_id=...]
        if request.values.get("app"):
            app_id = request.values.get("app")
            user_id = request.values.get("user", user_id)
            if (
                user_id != current_user.get_name()
                and not Permissions.PLUGIN_APPKEYS_ADMIN.can()
            ):
                return flask.abort(403)

            key = self._api_key_for_user_and_app_id(user_id, app_id)
            if not key:
                return flask.abort(404)

            return flask.jsonify(
                key=key.external(incl_key=credentials_checked_recently())
            )

        # GET ?all=true (admin only)
        if (
            request.values.get("all") in valid_boolean_trues
            and Permissions.PLUGIN_APPKEYS_ADMIN.can()
        ):
            keys = self._all_api_keys()

        else:
            keys = self._api_keys_for_user(user_id)

        return flask.jsonify(
            keys=list(
                map(lambda x: x.external(), keys),
            ),
            pending={
                x.user_token: x.external() for x in self._get_pending_by_user_id(user_id)
            },
        )

    def on_api_command(self, command, data):
        user_id = current_user.get_name()
        if not user_id:
            return flask.abort(403)

        if command == "revoke":
            api_key = data.get("key")

            if api_key:
                # deprecated key based revoke?
                from flask import request

                from octoprint.server.util import get_remote_address

                self._logger.warning(
                    "Deprecated key based revoke command sent to /api/plugin/appkeys by {}, should be migrated to use app id/user tuple".format(
                        get_remote_address(request)
                    )
                )

            else:
                # newer app/user based revoke?
                user = data.get("user", user_id)
                app = data.get("app")
                if not app:
                    return flask.abort(400, description="Need either app or key")

                api_key = self._api_key_for_user_and_app_id(user, app)

            if not api_key:
                return flask.abort(400, description="Need either app or key")

            if not Permissions.PLUGIN_APPKEYS_ADMIN.can():
                user_for_key = self._user_for_api_key(api_key)
                if user_for_key is None or user_for_key.user_id != user_id:
                    return flask.abort(403)

            ensure_credentials_checked_recently()

            self._delete_api_key(api_key)

        elif command == "generate":
            # manual generateKey
            app_name = data.get("app")
            if not app_name:
                return flask.abort(400)

            selected_user_id = data.get("user", user_id)
            if selected_user_id != user_id and not Permissions.PLUGIN_APPKEYS_ADMIN.can():
                return flask.abort(403)

            ensure_credentials_checked_recently()

            key = self._add_api_key(selected_user_id, app_name.strip())
            return flask.jsonify(user_id=selected_user_id, app_id=app_name, api_key=key)

        return NO_CONTENT

    ##~~ key validator hook

    def validate_api_key(self, api_key, *args, **kwargs):
        return self._user_for_api_key(api_key)

    ##~~ Helpers

    def _add_pending_decision(self, app_name, user_id=None):
        app_token = self._generate_key()
        user_token = self._generate_key()

        with self._pending_lock:
            self._remove_stale_pending()
            self._pending_decisions.append(
                PendingDecision(
                    app_name,
                    app_token,
                    user_id,
                    user_token,
                    timeout_callback=self._expire_pending,
                )
            )

        return app_token, user_token

    def _get_pending_by_app_token(self, app_token):
        result = []
        with self._pending_lock:
            self._remove_stale_pending()
            for data in self._pending_decisions:
                if data.app_token == app_token:
                    result.append(data)
        return result

    def _get_pending_by_user_id(self, user_id):
        result = []
        with self._pending_lock:
            self._remove_stale_pending()
            for data in self._pending_decisions:
                if data.user_id == user_id or data.user_id is None:
                    result.append(data)
        return result

    def _expire_pending(self, user_token):
        with self._pending_lock:
            len_before = len(self._pending_decisions)
            self._pending_decisions = list(
                filter(lambda x: x.user_token != user_token, self._pending_decisions)
            )
            len_after = len(self._pending_decisions)

            if len_after < len_before:
                self._plugin_manager.send_plugin_message(
                    self._identifier, {"type": "end_request", "user_token": user_token}
                )

    def _remove_stale_pending(self):
        with self._pending_lock:
            cutoff = time.monotonic() - CUTOFF_TIME
            len_before = len(self._pending_decisions)
            self._pending_decisions = list(
                filter(lambda x: x.created >= cutoff, self._pending_decisions)
            )
            len_after = len(self._pending_decisions)
            if len_after < len_before:
                self._logger.info(
                    "Deleted {} stale pending authorization requests".format(
                        len_before - len_after
                    )
                )

    def _set_decision(self, user_token, decision, user_id):
        with self._pending_lock:
            self._remove_stale_pending()
            for data in self._pending_decisions:
                if data.user_token == user_token and (
                    data.user_id == user_id or data.user_id is None
                ):
                    pending = data
                    break
            else:
                return False  # not found

        if decision:
            with self._ready_lock:
                self._ready_decisions.append(ReadyDecision.for_pending(pending, user_id))

        with self._pending_lock:
            self._pending_decisions = list(
                filter(lambda x: x.user_token != user_token, self._pending_decisions)
            )

        return True

    def _get_decision(self, app_token):
        self._remove_stale_pending()

        with self._ready_lock:
            for data in self._ready_decisions:
                if data.app_token == app_token:
                    decision = data
                    break
            else:
                return False  # not found

        api_key = self._add_api_key(decision.user_id, decision.app_id)

        with self._ready_lock:
            self._ready_decisions = list(
                filter(lambda x: x.app_token != app_token, self._ready_decisions)
            )

        return api_key

    def _add_api_key(self, user_id, app_name):
        with self._keys_lock:
            for key in self._keys[user_id]:
                if key.app_id.lower() == app_name.lower():
                    return key.api_key

            key = ActiveKey(app_name, self._generate_key(), user_id)
            self._keys[user_id].append(key)
            self._save_keys()
            return key.api_key

    def _delete_api_key(self, api_key):
        if isinstance(api_key, ActiveKey):
            api_key = api_key.api_key

        with self._keys_lock:
            for user_id, data in self._keys.items():
                self._keys[user_id] = list(filter(lambda x: x.api_key != api_key, data))
            self._save_keys()

    def _user_for_api_key(self, api_key):
        if isinstance(api_key, ActiveKey):
            api_key = api_key.api_key

        with self._keys_lock:
            for user_id, data in self._keys.items():
                if any(filter(lambda x: x.api_key == api_key, data)):
                    return self._user_manager.find_user(userid=user_id)
        return None

    def _api_keys_for_user(self, user_id):
        with self._keys_lock:
            return self._keys[user_id]

    def _all_api_keys(self):
        with self._keys_lock:
            result = []
            for keys in self._keys.values():
                result += keys
        return result

    def _api_key_for_user_and_app_id(self, user_id, app_id):
        with self._keys_lock:
            if user_id not in self._keys:
                return None

            for key in self._keys[user_id]:
                if key.app_id.lower() == app_id.lower():
                    return key

        return None

    def _generate_key(self):
        return generate_api_key()

    def _load_keys(self):
        with self._keys_lock:
            if not os.path.exists(self._key_path):
                return

            try:
                persisted = yaml.load_from_file(path=self._key_path)
            except Exception:
                self._logger.exception(
                    f"Could not load application keys from {self._key_path}"
                )
                return

            if not isinstance(persisted, dict):
                return

            keys = defaultdict(list)
            for user_id, persisted_keys in persisted.items():
                keys[user_id] = [
                    ActiveKey.for_internal(x, user_id) for x in persisted_keys
                ]
            self._keys = keys

    def _save_keys(self):
        with self._keys_lock:
            to_persist = {}
            for user_id, keys in self._keys.items():
                to_persist[user_id] = [x.internal() for x in keys]

            try:
                with atomic_write(self._key_path, mode="wt") as f:
                    yaml.save_to_file(to_persist, file=f)
            except Exception:
                self._logger.exception(
                    f"Could not write application keys to {self._key_path}"
                )


__plugin_name__ = "Application Keys Plugin"
__plugin_description__ = (
    "Implements a workflow for third party clients to obtain API keys"
)
__plugin_author__ = "Gina Häußge, Aldo Hoeben"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin third party clients will no longer be able to "
    "obtain an API key without you manually copy-pasting it."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_implementation__ = AppKeysPlugin()
__plugin_hooks__ = {
    "octoprint.accesscontrol.keyvalidator": __plugin_implementation__.validate_api_key,
    "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
}



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import concurrent.futures
import hashlib
import logging
import time
from urllib.parse import urlencode

import requests
from flask_babel import gettext

import octoprint.plugin
from octoprint.events import Events
from octoprint.util import RepeatedTimer
from octoprint.util.version import get_octoprint_version_string

TRACKING_URL = "https://tracking.octoprint.org/track/{id}/{event}/"


# noinspection PyMissingConstructor
class TrackingPlugin(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.EnvironmentDetectionPlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.ShutdownPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.WizardPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.SimpleApiPlugin,
):
    def __init__(self):
        self._environment = None
        self._throttle_state = None
        self._helpers_get_throttle_state = None
        self._helpers_get_unlocked_achievements = None
        self._printer_connection_parameters = None
        self._url = None
        self._ping_worker = None
        self._pong_worker = None
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        self._record_next_firmware_info = False

        self._startup_time = time.monotonic()

    def initialize(self):
        self._init_id()

    ##~~ SettingsPlugin

    def get_settings_defaults(self):
        return {
            "enabled": None,
            "unique_id": None,
            "server": TRACKING_URL,
            "ping": 15 * 60,
            "pong": 24 * 60 * 60,
            "events": {
                "pong": True,
                "startup": True,
                "printjob": True,
                "commerror": True,
                "plugin": True,
                "update": True,
                "printer": True,
                "printer_safety_check": True,
                "throttled": True,
                "achievements": True,
                "slicing": True,
                "webui_load": True,
            },
        }

    def get_settings_restricted_paths(self):
        return {
            "admin": [["enabled"], ["unique_id"], ["events"]],
            "never": [["server"], ["ping"]],
        }

    def on_settings_save(self, data):
        enabled = self._settings.get(["enabled"])

        octoprint.plugin.SettingsPlugin.on_settings_save(self, data)

        if enabled is None and self._settings.get(["enabled"]):
            # tracking was just enabled, let's start up tracking
            self._start_tracking()

    ##~~ SimpleApiPlugin

    def get_api_commands(self):
        return {"track": ["event", "payload"]}

    def on_api_command(self, command, data):
        if command != "track":
            return

        event = data.pop("event")
        if event == "webui_load":
            payload = data.get("payload", {})
            browser_name = payload.get("browser_name")
            browser_version = payload.get("browser_version")
            os_name = payload.get("os_name")
            os_version = payload.get("os_version")

            if all((browser_name, browser_version, os_name, os_version)):
                self._track_webui_load(browser_name, browser_version, os_name, os_version)

    ##~~ EnvironmentDetectionPlugin

    def on_environment_detected(self, environment, *args, **kwargs):
        self._environment = environment

    ##~~ StartupPlugin

    def on_after_startup(self):
        self._start_tracking()

    ##~~ ShutdownPlugin

    def on_shutdown(self):
        if not self._settings.get_boolean(["enabled"]):
            return
        self._track_shutdown()
        self._executor.shutdown(wait=True)

    ##~~ EventHandlerPlugin

    # noinspection PyUnresolvedReferences
    def on_event(self, event, payload):
        if not self._settings.get_boolean(["enabled"]):
            return

        if event in (
            Events.PRINT_STARTED,
            Events.PRINT_DONE,
            Events.PRINT_FAILED,
            Events.PRINT_CANCELLED,
        ):
            self._track_printjob_event(event, payload)

        elif event in (Events.ERROR,):
            self._track_commerror_event(event, payload)

        elif event in (Events.CONNECTED,):
            self._printer_connection_parameters = {
                "port": payload["port"],
                "baudrate": payload["baudrate"],
            }
            self._record_next_firmware_info = True

        elif event in (Events.FIRMWARE_DATA,) and self._record_next_firmware_info:
            self._record_next_firmware_info = False
            self._track_printer_event(event, payload)

        elif event in (Events.SLICING_STARTED,):
            self._track_slicing_event(event, payload)

        elif hasattr(Events, "PLUGIN_PLUGINMANAGER_INSTALL_PLUGIN") and event in (
            Events.PLUGIN_PLUGINMANAGER_INSTALL_PLUGIN,
            Events.PLUGIN_PLUGINMANAGER_UNINSTALL_PLUGIN,
            Events.PLUGIN_PLUGINMANAGER_ENABLE_PLUGIN,
            Events.PLUGIN_PLUGINMANAGER_DISABLE_PLUGIN,
        ):
            self._track_plugin_event(event, payload)

        elif hasattr(Events, "PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED") and event in (
            Events.PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED,
            Events.PLUGIN_SOFTWAREUPDATE_UPDATE_FAILED,
        ):
            self._track_update_event(event, payload)

        elif hasattr(Events, "PLUGIN_PI_SUPPORT_THROTTLE_STATE") and event in (
            Events.PLUGIN_PI_SUPPORT_THROTTLE_STATE,
        ):
            self._throttle_state = payload
            self._track_throttle_event(event, payload)

        elif hasattr(Events, "PLUGIN_PRINTER_SAFETY_CHECK_WARNING") and event in (
            Events.PLUGIN_PRINTER_SAFETY_CHECK_WARNING,
        ):
            self._track_printer_safety_event(event, payload)

        elif hasattr(Events, "PLUGIN_ACHIEVEMENTS_ACHIEVEMENT_UNLOCKED") and event in (
            Events.PLUGIN_ACHIEVEMENTS_ACHIEVEMENT_UNLOCKED,
        ):
            self._track_achievement_unlocked_event(event, payload)

    ##~~ TemplatePlugin

    def get_template_configs(self):
        return [
            {
                "type": "settings",
                "name": gettext("Anonymous Usage Tracking"),
                "template": "tracking_settings.jinja2",
                "custom_bindings": False,
            },
            {
                "type": "wizard",
                "name": gettext("Anonymous Usage Tracking"),
                "template": "tracking_wizard.jinja2",
                "custom_bindings": True,
                "mandatory": True,
            },
        ]

    ##~~ AssetPlugin

    def get_assets(self):
        return {"js": ["js/usage.js"], "clientjs": ["clientjs/usage.js"]}

    ##~~ WizardPlugin

    def is_wizard_required(self):
        return self._settings.get(["enabled"]) is None

    ##~~ helpers

    def _init_id(self):
        if not self._settings.get(["unique_id"]):
            import uuid

            self._settings.set(["unique_id"], str(uuid.uuid4()))
            self._settings.save()

    def _start_tracking(self):
        if not self._settings.get_boolean(["enabled"]):
            return

        if self._helpers_get_throttle_state is None:
            # cautiously look for the get_throttled helper from pi_support
            pi_helper = self._plugin_manager.get_helpers("pi_support", "get_throttled")
            if pi_helper and "get_throttled" in pi_helper:
                self._helpers_get_throttle_state = pi_helper["get_throttled"]

        if self._helpers_get_unlocked_achievements is None:
            # cautiously look for the get_unlocked_achievements helper from achievements
            achievements_helper = self._plugin_manager.get_helpers(
                "achievements", "get_unlocked_achievements"
            )
            if achievements_helper and "get_unlocked_achievements" in achievements_helper:
                self._helpers_get_unlocked_achievements = achievements_helper[
                    "get_unlocked_achievements"
                ]

        if self._ping_worker is None:
            ping_interval = self._settings.get_int(["ping"])
            if ping_interval:
                self._ping_worker = RepeatedTimer(
                    ping_interval, self._track_ping, run_first=True
                )
                self._ping_worker.start()

        if self._pong_worker is None:
            pong_interval = self._settings.get(["pong"])
            if pong_interval:
                self._pong_worker = RepeatedTimer(
                    pong_interval, self._track_pong, run_first=True
                )
                self._pong_worker.start()

        # now that we have everything set up, phone home.
        self._track_startup()

    def _track_ping(self):
        if not self._settings.get_boolean(["enabled"]):
            return

        uptime = int(time.monotonic() - self._startup_time)
        printer_state = self._printer.get_state_id()
        self._track("ping", octoprint_uptime=uptime, printer_state=printer_state)

    def _track_pong(self):
        if not self._settings.get_boolean(["events", "pong"]):
            return

        payload = self._get_environment_payload()

        plugins = self._plugin_manager.enabled_plugins
        plugins_thirdparty = [plugin for plugin in plugins.values() if not plugin.bundled]
        payload["plugins"] = ",".join(
            map(
                lambda x: "{}:{}".format(
                    x.key.lower(), x.version.lower() if x.version else "?"
                ),
                plugins_thirdparty,
            )
        )

        if self._helpers_get_unlocked_achievements and self._settings.get_boolean(
            ["events", "achievements"]
        ):
            payload["achievements"] = ",".join(
                map(
                    lambda x: x.key.lower(),
                    self._helpers_get_unlocked_achievements(),
                )
            )

        self._track("pong", body=True, **payload)

    def _track_startup(self):
        if not self._settings.get_boolean(["events", "startup"]):
            return

        payload = self._get_environment_payload()

        self._track("startup", **payload)

    def _track_shutdown(self):
        if not self._settings.get_boolean(["enabled"]):
            return

        if not self._settings.get_boolean(["events", "startup"]):
            return

        self._track("shutdown")

    def _track_webui_load(self, browser_name, browser_version, os_name, os_version):
        if not self._settings.get_boolean(["events", "webui_load"]):
            return

        self._track(
            "webui_load",
            browser_name=browser_name,
            browser_version=browser_version,
            os_name=os_name,
            os_version=os_version,
        )

    def _track_plugin_event(self, event, payload):
        if not self._settings.get_boolean(["events", "plugin"]):
            return

        if event.endswith("_install_plugin"):
            self._track(
                "install_plugin",
                plugin=payload.get("id"),
                plugin_version=payload.get("version"),
            )
        elif event.endswith("_uninstall_plugin"):
            self._track(
                "uninstall_plugin",
                plugin=payload.get("id"),
                plugin_version=payload.get("version"),
            )
        elif event.endswith("_enable_plugin"):
            self._track(
                "enable_plugin",
                plugin=payload.get("id"),
                plugin_version=payload.get("version"),
            )
        elif event.endswith("_disable_plugin"):
            self._track(
                "disable_plugin",
                plugin=payload.get("id"),
                plugin_version=payload.get("version"),
            )

    def _track_update_event(self, event, payload):
        if not self._settings.get_boolean(["events", "update"]):
            return

        if event.endswith("_update_succeeded"):
            self._track(
                "update_successful",
                target=payload.get("target"),
                from_version=payload.get("from_version"),
                to_version=payload.get("to_version"),
            )
        elif event.endswith("_update_failed"):
            self._track(
                "update_failed",
                target=payload.get("target"),
                from_version=payload.get("from_version"),
                to_version=payload.get("to_version"),
            )

    def _track_throttle_event(self, event, payload):
        if not self._settings.get_boolean(["events", "throttled"]):
            return

        args = {
            "throttled_now": payload["current_issue"],
            "throttled_past": payload["past_issue"],
            "throttled_mask": payload["raw_value"],
            "throttled_voltage_now": payload["current_undervoltage"],
            "throttled_voltage_past": payload["past_undervoltage"],
            "throttled_overheat_now": payload["current_overheat"],
            "throttled_overheat_past": payload["past_overheat"],
        }

        if payload["current_issue"]:
            track_event = "system_throttled"
        else:
            track_event = "system_unthrottled"

        if track_event is not None:
            self._track(track_event, **args)

    def _track_commerror_event(self, event, payload):
        if not self._settings.get_boolean(["events", "commerror"]):
            return

        if "reason" not in payload or "error" not in payload:
            return

        track_event = "commerror_{}".format(payload["reason"])
        args = {"commerror_text": payload["error"]}

        if callable(self._helpers_get_throttle_state):
            try:
                throttle_state = self._helpers_get_throttle_state(run_now=True)
                if throttle_state and (
                    throttle_state.get("current_issue", False)
                    or throttle_state.get("past_issue", False)
                ):
                    args["throttled_now"] = throttle_state["current_issue"]
                    args["throttled_past"] = throttle_state["past_issue"]
                    args["throttled_mask"] = throttle_state["raw_value"]
            except Exception:
                # ignored
                pass

        self._track(track_event, **args)

    def _track_printjob_event(self, event, payload):
        if not self._settings.get_boolean(["events", "printjob"]):
            return

        unique_id = self._settings.get(["unique_id"])
        if not unique_id:
            return

        sha = hashlib.sha1()
        sha.update(payload.get("path").encode("utf-8"))
        sha.update(unique_id.encode("utf-8"))

        track_event = None
        args = {"origin": payload.get("origin"), "file": sha.hexdigest()}

        if event == Events.PRINT_STARTED:
            track_event = "print_started"
        elif event == Events.PRINT_DONE:
            try:
                elapsed = int(payload.get("time", 0))
                if elapsed:
                    args["elapsed"] = elapsed
            except (ValueError, TypeError):
                pass
            track_event = "print_done"
        elif event == Events.PRINT_FAILED:
            try:
                elapsed = int(payload.get("time", 0))
                if elapsed:
                    args["elapsed"] = elapsed
            except (ValueError, TypeError):
                pass
            args["reason"] = payload.get("reason", "unknown")

            if "error" in payload and self._settings.get_boolean(["events", "commerror"]):
                args["commerror_text"] = payload["error"]

            track_event = "print_failed"
        elif event == Events.PRINT_CANCELLED:
            try:
                elapsed = int(payload.get("time", 0))
                if elapsed:
                    args["elapsed"] = elapsed
            except (ValueError, TypeError):
                pass
            track_event = "print_cancelled"

        if callable(self._helpers_get_throttle_state):
            try:
                throttle_state = self._helpers_get_throttle_state(run_now=True)
                if throttle_state and (
                    throttle_state.get("current_issue", False)
                    or throttle_state.get("past_issue", False)
                ):
                    args["throttled_now"] = throttle_state["current_issue"]
                    args["throttled_past"] = throttle_state["past_issue"]
                    args["throttled_mask"] = throttle_state["raw_value"]
            except Exception:
                # ignored
                pass

        if track_event is not None:
            self._track(track_event, **args)

    def _track_printer_event(self, event, payload):
        if not self._settings.get_boolean(["events", "printer"]):
            return

        if event in (Events.FIRMWARE_DATA,):
            args = {"firmware_name": payload["name"]}
            if self._printer_connection_parameters:
                args["printer_port"] = self._printer_connection_parameters["port"]
                args["printer_baudrate"] = self._printer_connection_parameters["baudrate"]
            self._track("printer_connected", **args)

    def _track_printer_safety_event(self, event, payload):
        if not self._settings.get_boolean(["events", "printer_safety_check"]):
            return

        self._track(
            "printer_safety_warning",
            printer_safety_warning_type=payload.get("warning_type", "unknown"),
            printer_safety_check_name=payload.get("check_name", "unknown"),
        )

    def _track_achievement_unlocked_event(self, event, payload):
        if not self._settings.get_boolean(["events", "achievements"]):
            return

        self._track(
            "achievement_unlocked",
            achievement=payload.get("key", "unknown"),
        )

    def _track_slicing_event(self, event, payload):
        if not self._settings.get_boolean(["events", "slicing"]):
            return

        self._track("slicing_started", slicer=payload.get(b"slicer", "unknown"))

    def _track(self, event, **kwargs):
        if not self._settings.get_boolean(["enabled"]):
            return

        self._executor.submit(self._do_track, event, **kwargs)

    def _do_track(self, event, body=False, **kwargs):
        if not self._connectivity_checker.online:
            return

        if not self._settings.get_boolean(["enabled"]):
            return

        unique_id = self._settings.get(["unique_id"])
        if not unique_id:
            return

        server = self._settings.get(["server"])
        url = server.format(id=unique_id, event=event)
        # Don't print the URL or UUID! That would expose the UUID in forums/tickets
        # if pasted. It's okay for the user to know their uuid, but it shouldn't be shared.

        headers = {"User-Agent": f"OctoPrint/{get_octoprint_version_string()}"}
        try:
            params = urlencode(kwargs, doseq=True).replace("+", "%20")

            if body:
                requests.post(url, data=params, timeout=3.05, headers=headers)
            else:
                requests.get(url, params=params, timeout=3.05, headers=headers)

            self._logger.info(f"Sent tracking event {event}, payload: {kwargs!r}")
        except Exception:
            if self._logger.isEnabledFor(logging.DEBUG):
                self._logger.exception(
                    "Error while sending event to anonymous usage tracking"
                )
            else:
                pass

    def _get_environment_payload(self):
        payload = {
            "version": get_octoprint_version_string(),
            "os": self._environment["os"]["id"],
            "bits": self._environment["os"]["bits"],
            "python": self._environment["python"]["version"],
            "pip": self._environment["python"]["pip"],
            "cores": self._environment["hardware"]["cores"],
            "freq": self._environment["hardware"]["freq"],
            "ram": self._environment["hardware"]["ram"],
        }

        if (
            "plugins" in self._environment
            and "pi_support" in self._environment["plugins"]
        ):
            payload["pi_model"] = self._environment["plugins"]["pi_support"]["model"]

            if "octopi_version" in self._environment["plugins"]["pi_support"]:
                payload["octopi_version"] = self._environment["plugins"]["pi_support"][
                    "octopi_version"
                ]
            if "octopiuptodate_build" in self._environment["plugins"]["pi_support"]:
                payload["octopiuptodate_build"] = self._environment["plugins"][
                    "pi_support"
                ]["octopiuptodate_build"]

        return payload


__plugin_name__ = "Anonymous Usage Tracking"
__plugin_description__ = (
    "Anonymous version and usage tracking, see homepage for details on what gets tracked"
)
__plugin_url__ = "https://tracking.octoprint.org"
__plugin_author__ = "Gina Häußge"
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=3.7,<4"

__plugin_implementation__ = TrackingPlugin()



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import List, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class AlwaysDetectNeverEnum(str, Enum):
    always = "always"
    detect = "detect"
    never = "never"


class InfoWarnNeverEnum(str, Enum):
    info = "info"
    warn = "warn"
    never = "never"


@with_attrs_docs
class SerialTimeoutConfig(BaseModel):
    detectionFirst: float = 10.0
    detectionConsecutive: float = 2.0

    connection: float = 10.0
    """Timeout for waiting to establish a connection with the selected port, in seconds"""

    communication: float = 30.0
    """Timeout during serial communication, in seconds"""

    communicationBusy: float = 3.0
    """Timeout during serial communication when busy protocol support is detected, in seconds"""

    temperature: float = 5.0
    """Timeout after which to query temperature when no target is set"""

    temperatureTargetSet: float = 2.0
    """Timeout after which to query temperature when a target is set"""

    temperatureAutoreport: float = 2.0

    sdStatus: float = 1.0
    """Timeout after which to query the SD status while SD printing"""

    sdStatusAutoreport: float = 1.0
    posAutoreport: float = 5.0
    resendOk: float = 0.5
    baudrateDetectionPause: float = 1.0
    positionLogWait: float = 10.0


@with_attrs_docs
class SerialMaxTimeouts(BaseModel):
    idle: int = 2
    """Max. timeouts when the printer is idle"""

    printing: int = 5
    """Max. timeouts when the printer is printing"""

    long: int = 5
    """Max. timeouts when a long running command is active"""


@with_attrs_docs
class SerialCapabilities(BaseModel):
    autoreport_temp: bool = True
    """Whether to enable temperature autoreport in the firmware if its support is detected"""

    autoreport_sdstatus: bool = True
    """Whether to enable SD printing autoreport in the firmware if its support is detected"""

    autoreport_pos: bool = True
    """Whether to enable position autoreport in the firmware if its support is detected"""

    busy_protocol: bool = True
    """Whether to shorten the communication timeout if the firmware seems to support the busy protocol"""

    emergency_parser: bool = True
    """Whether to send emergency commands out of band if the firmware seems to support the emergency parser"""

    extended_m20: bool = True
    """Whether to request extended M20 (file list) output from the firmware if its support is detected"""

    lfn_write: bool = True
    """Whether to enable long filename support for SD card writes if the firmware reports support for it"""


@with_attrs_docs
class SerialConfig(BaseModel):
    port: Optional[str] = None
    """The default port to use to connect to the printer. If unset or set to `AUTO`, the port will be auto-detected."""

    baudrate: Optional[int] = None
    """The default baudrate to use to connect to the printer. If unset or set to 0, the baudrate will be auto-detected."""

    exclusive: bool = True
    """Whether to request the serial port exclusively or not"""

    lowLatency: bool = False
    """Whether to request low latency mode on the serial port or not"""

    autoconnect: bool = False
    """Whether to try to automatically connect to the printer on startup or not"""

    autorefresh: bool = True
    """Whether to automatically refresh the port list while no connection is established"""

    autorefreshInterval: int = 1
    """Interval in seconds at which to refresh the port list while no connection is established"""

    log: bool = False
    """Whether to log whole communication to serial.log (warning: might decrease performance)"""

    timeout: SerialTimeoutConfig = SerialTimeoutConfig()
    """Timeouts used for the serial connection to the printer, you might want to adjust these if you are experiencing connection problems"""

    maxCommunicationTimeouts: SerialMaxTimeouts = SerialMaxTimeouts()

    maxWritePasses: int = 5
    """Maximum number of write attempts to serial during which nothing can be written before the communication with the printer is considered dead and OctoPrint will disconnect with an error"""

    additionalPorts: List[str] = []
    """Use this to define additional patterns to consider for serial port listing. Must be a valid ["glob" pattern](http://docs.python.org/3/library/glob.html)"""

    additionalBaudrates: List[int] = []
    """Use this to define additional baud rates to offer for connecting to serial ports. Must be a valid integer"""

    blacklistedPorts: List[str] = []
    blacklistedBaudrates: List[int] = []

    longRunningCommands: List[str] = [
        "G4",
        "G28",
        "G29",
        "G30",
        "G32",
        "M400",
        "M226",
        "M600",
    ]
    """Commands which are known to take a long time to be acknowledged by the firmware, e.g. homing, dwelling, auto leveling etc."""

    blockedCommands: List[str] = ["M0", "M1"]
    """Commands which should not be sent to the printer, e.g. because they are known to block serial communication until physical interaction with the printer as is the case on most firmwares with the default M0 and M1."""

    ignoredCommands: List[str] = []
    """Commands which should not be sent to the printer and just silently ignored. An example of when you may wish to use this could be useful if you wish to manually change a filament on M600, by using that as a Pausing command."""

    pausingCommands: List[str] = ["M0", "M1", "M25"]
    """Commands which should cause OctoPrint to pause any ongoing prints."""

    emergencyCommands: List[str] = ["M112", "M108", "M410"]

    checksumRequiringCommands: List[str] = ["M110"]
    """Commands which need to always be send with a checksum."""

    helloCommand: str = "M110 N0"
    """Command to send in order to initiate a handshake with the printer."""

    disconnectOnErrors: bool = True
    """Whether to disconnect from the printer on errors or not."""

    ignoreErrorsFromFirmware: bool = False
    """Whether to completely ignore errors from the firmware or not."""

    terminalLogSize: int = 20
    lastLineBufferSize: int = 50

    logResends: bool = True
    """Whether to log resends to octoprint.log or not. Invaluable debug tool without performance impact, leave on if possible please."""

    supportResendsWithoutOk: AlwaysDetectNeverEnum = "detect"
    """Whether to support resends without follow-up ok or not."""

    logPositionOnPause: bool = True
    logPositionOnCancel: bool = False
    abortHeatupOnCancel: bool = True

    waitForStartOnConnect: bool = False
    """Whether OctoPrint should wait for the `start` response from the printer before trying to send commands during connect."""

    waitToLoadSdFileList: bool = True
    """Specifies whether OctoPrint should wait to load the SD card file list until the first firmware capability report is processed."""

    alwaysSendChecksum: bool = False
    """Specifies whether OctoPrint should send linenumber + checksum with every printer command. Needed for successful communication with Repetier firmware."""

    neverSendChecksum: bool = False

    sendChecksumWithUnknownCommands: bool = False
    r"""Specifies whether OctoPrint should also send linenumber + checksum with commands that are *not* detected as valid GCODE (as in, they do not match the regular expression `^\s*([GM]\d+|T)`)."""

    unknownCommandsNeedAck: bool = False
    r"""Specifies whether OctoPrint should also use up acknowledgments (`ok`) for commands that are *not* detected as valid GCODE (as in, they do not match the regular expression `^\s*([GM]\d+|T)`)."""

    sdRelativePath: bool = False
    """Specifies whether firmware expects relative paths for selecting SD files."""

    sdAlwaysAvailable: bool = False
    """Whether to always assume that an SD card is present in the printer. Needed by some firmwares which don't report the SD card status properly."""

    sdLowerCase: bool = False
    sdCancelCommand: str = "M25"
    maxNotSdPrinting: int = 2
    swallowOkAfterResend: bool = True

    repetierTargetTemp: bool = False
    """Whether the printer sends repetier style target temperatures in the format `TargetExtr0:<temperature>` instead of attaching that information to the regular `M105` responses."""

    externalHeatupDetection: bool = True
    """Whether to enable external heatup detection (to detect heatup triggered e.g. through the printer's LCD panel or while printing from SD) or not. Causes issues with Repetier's "first ok then response" approach to communication, so disable for printers running Repetier firmware."""

    supportWait: bool = True

    ignoreIdenticalResends: bool = False
    """Whether to ignore identical resends from the printer (true, repetier) or not (false)."""

    identicalResendsCountdown: int = 7
    """If `ignoreIdenticalResends` is true, how many consecutive identical resends to ignore."""

    supportFAsCommand: bool = False
    """Whether to support `F` on its own as a valid GCODE command (true) or not (false)."""

    firmwareDetection: bool = True
    """Whether to attempt to auto detect the firmware of the printer and adjust settings accordingly (true) or not and rely on manual configuration (false)."""

    blockWhileDwelling: bool = False
    """Whether to block all sending to the printer while a G4 (dwell) command is active (true, repetier) or not (false)."""

    useParityWorkaround: AlwaysDetectNeverEnum = "detect"
    maxConsecutiveResends: int = 10
    sendM112OnError: bool = True
    disableSdPrintingDetection: bool = False
    ackMax: int = 1
    sanityCheckTools: bool = True
    notifySuppressedCommands: InfoWarnNeverEnum = "warn"
    capabilities: SerialCapabilities = SerialCapabilities()

    resendRatioThreshold: int = 10
    """Percentage of resend requests among all sent lines that should be considered critical."""

    resendRatioStart: int = 100
    ignoreEmptyPorts: bool = False

    encoding: str = "ascii"
    """Encoding to use when talking to a machine. `ascii` limits access to characters 0-127, `latin_1` enables access to the "extended" ascii characters 0-255. Other values can be used, see [Python's standard encodings](https://docs.python.org/3/library/codecs.html#standard-encodings)."""

    enableShutdownActionCommand: bool = False
    """Whether to enable support for the shutdown action command, allowing the printer to shut down OctoPrint and the system it's running on."""

    # command specific flags
    triggerOkForM29: bool = True
    """Whether to automatically trigger an ok for `M29` (a lot of versions of this command are buggy and the response skips on the ok)."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class TemperatureProfile(BaseModel):
    name: str
    """Name of the profile."""

    extruder: int
    """Hotend temperature to set with the profile."""

    bed: int
    """Bed temperature to set with the profile."""


@with_attrs_docs
class TemperatureConfig(BaseModel):
    profiles: List[TemperatureProfile] = [
        TemperatureProfile(name="ABS", extruder=210, bed=100),
        TemperatureProfile(name="PLA", extruder=180, bed=60),
    ]
    """Temperature profiles to offer in the UI for quick pre-heating."""

    cutoff: int = 30
    """Cut off time for the temperature data, in minutes."""

    sendAutomatically: bool = False
    """Whether to send new temperature settings made in the UI automatically."""

    sendAutomaticallyAfter: int = 1
    """After what time to send the new temperature settings automatically, in seconds."""



import os
import re

from pydantic import BaseModel

ROMAN_NUMERAL = re.compile(
    "^M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$", re.IGNORECASE
)


class Achievement(BaseModel):
    key: str = ""
    name: str = ""
    description: str = ""
    hidden: bool = False
    nag: bool = False

    @property
    def icon(self):
        key = self.key
        if "_" in key and ROMAN_NUMERAL.match(key.rsplit("_", 1)[1]):
            key = key.rsplit("_", 1)[0]

        if os.path.exists(
            os.path.join(os.path.dirname(__file__), "static", "img", f"{key}.svg")
        ):
            return f"{key}"
        else:
            return "trophy"


class AchievementsMetaClass(type):
    achievements = {}
    key_to_attr = {}

    def __new__(mcs, name, bases, args):
        cls = type.__new__(mcs, name, bases, args)

        for key, value in args.items():
            if isinstance(value, Achievement):
                value.key = key.lower()
                mcs.achievements[key] = value
                mcs.key_to_attr[key.lower()] = key

        return cls

    def all(cls):
        return cls.achievements.values()

    def get(cls, key):
        attr = cls.key_to_attr.get(key)
        if not attr:
            return None

        return cls.achievements.get(attr)


class Achievements(metaclass=AchievementsMetaClass):
    ## basics

    THE_WIZARD = Achievement(
        name="The Wizard",
        description="Complete the first run setup wizard.",
    )

    ONE_SMALL_STEP_FOR_MAN = Achievement(
        name="That's One Small Step For Man", description="Finish your first print."
    )

    ADVENTURER = Achievement(
        name="The Adventurer",
        description="Install a plugin.",
    )

    TINKERER = Achievement(
        name="The Tinkerer",
        description="Install a plugin from a URL.",
    )

    BETTER_SAFE_THAN_SORRY = Achievement(
        name="Better Safe Than Sorry",
        description="Create a backup.",
    )

    ## print duration

    HALF_MARATHON = Achievement(
        name="Half Marathon",
        description="Finish a print that took longer than 12 hours.",
        nag=True,
    )

    MARATHON = Achievement(
        name="Marathon",
        description="Finish a print that took longer than 24 hours.",
        nag=True,
    )

    SPRINT = Achievement(
        name="Sprint",
        description="Finish a print that took less than 10 minutes.",
    )

    ACHIEVEMENT_NOT_FOUND = Achievement(
        name="Achievement Not Found",
        description="Reach a total print duration of 404 hours",
        hidden=True,
        nag=True,
    )

    ## print count

    CANT_GET_ENOUGH = Achievement(
        name="Can't Get Enough",
        description="Finish 10 prints in one day.",
        nag=True,
    )

    THE_MANUFACTURER_I = Achievement(
        name="The Manufacturer", description="Finish 10 prints.", nag=True
    )

    THE_MANUFACTURER_II = Achievement(
        name="The Manufacturer II",
        description="Finish 100 prints.",
        hidden=True,
        nag=True,
    )

    THE_MANUFACTURER_III = Achievement(
        name="The Manufacturer III",
        description="Finish 1000 prints.",
        hidden=True,
        nag=True,
    )

    ## date or date range specific

    HAPPY_BIRTHDAY_FOOSEL = Achievement(
        name="Happy Birthday, foosel",
        description="Start a print on foosel's birthday, March 21st.",
        hidden=True,
    )

    HAPPY_BIRTHDAY_OCTOPRINT = Achievement(
        name="Happy Birthday, OctoPrint",
        description="Start a print on OctoPrint's birthday, December 25th.",
        hidden=True,
    )

    SPOOKY = Achievement(
        name="Spooky",
        description="Start a print on Halloween, October 31st.",
        hidden=True,
    )

    SANTAS_LITTLE_HELPER = Achievement(
        name="Santa's Little Helper",
        description="Start a print between December 1st and December 24th.",
        hidden=True,
    )

    ## weekday specific

    TGIF = Achievement(
        name="TGIF",
        description="Start a print on a Friday.",
    )

    WEEKEND_WARRIOR = Achievement(
        name="Weekend Warrior",
        description="Start prints on four consecutive weekends.",
    )

    ## time specific

    EARLY_BIRD = Achievement(
        name="Early Bird",
        description="Start a print between 03:00 and 07:00.",
        hidden=True,
    )

    NIGHT_OWL = Achievement(
        name="Night Owl",
        description="Start a print between 23:00 and 03:00.",
        hidden=True,
    )

    ## file management

    CLEAN_HOUSE_I = Achievement(
        name="Clean House",
        description="Delete 100 files.",
        hidden=True,
    )

    CLEAN_HOUSE_II = Achievement(
        name="Clean House II",
        description="Delete 500 files.",
        hidden=True,
    )

    CLEAN_HOUSE_III = Achievement(
        name="Clean House III",
        description="Delete 1000 files.",
        hidden=True,
    )

    HEAVY_CHONKER = Achievement(
        name="Heavy Chonker",
        description="Upload a GCODE file larger than 500MB.",
    )

    THE_COLLECTOR_I = Achievement(
        name="The Collector",
        description="Upload 100 files.",
        hidden=True,
        nag=True,
    )

    THE_COLLECTOR_II = Achievement(
        name="The Collector II",
        description="Upload 500 files.",
        hidden=True,
        nag=True,
    )

    THE_COLLECTOR_III = Achievement(
        name="The Collector III",
        description="Upload 1000 files.",
        hidden=True,
        nag=True,
    )

    THE_ORGANIZER = Achievement(
        name="The Organizer",
        description="Create a folder.",
        hidden=True,
    )

    ## mishaps

    ALL_BEGINNINGS_ARE_HARD = Achievement(
        name="All Beginnings Are Hard", description="Cancel your first print."
    )

    ONE_OF_THOSE_DAYS = Achievement(
        name="Must Be One Of Those Days",
        description="Cancel ten consecutive prints on the same day.",
        hidden=True,
    )

    SO_CLOSE = Achievement(
        name="So Close",
        description="Cancel a print job at 95% progress or more.",
        hidden=True,
    )

    ## misc

    CROSSOVER_EPISODE = Achievement(
        name="What Is This, A Crossover Episode?",
        description="Connect to a printer running Klipper.",
        hidden=True,
    )

    HANG_IN_THERE = Achievement(
        name="Hang In There!",
        description="Pause the same print ten times.",
        hidden=True,
    )

    MASS_PRODUCTION = Achievement(
        name="Mass Production",
        description="Finish a print of the same file five times in a row.",
        hidden=True,
        nag=True,
    )

    WHAT_COULD_POSSIBLY_GO_WRONG = Achievement(
        name="What Could Possibly Go Wrong?",
        description="Start a print with an active undervoltage issue.",
        hidden=True,
    )


if __name__ == "__main__":
    import json

    achievements = Achievements.all()

    print("const ACHIEVEMENTS = ", end="")
    print(
        json.dumps(
            {
                a.key: {"name": a.name, "hidden": a.hidden}
                for a in sorted(achievements, key=lambda x: x.key)
            },
            indent=2,
        )
    )
